"""
    mave2imap: This module provide means for infering and mapping interface
    hotspots based on results from MAVE (Multiplexed Assays of Variant
    Effects).

    The main function of this file corresponds to the entry point of mave2imap
    executable script
"""

# Modules import
# OS related
import os
import sys
from subprocess import run as sps_run
# Utils
from datetime import datetime
import argparse
import configparser
import warnings
warnings.filterwarnings("ignore")

# Global variables
FINAL_INIT_FN = 'final_init.ini'
fastq_files = ['f1', 'f2', 'f3', 'f4']


def init_parse(ini_f):
    """
        Recovers the parameters contained in initialization file (provide by -i or --ini argument)
    """

    init_dict = {}

    config = configparser.ConfigParser(interpolation=configparser.ExtendedInterpolation())
    config.read(ini_f)
    # print("****  START CONFIG FILE CONTENT ****\n")
    # print(open(ini_f, encoding="utf-8").read())
    # print("****  END CONFIG FILE CONTENT ****\n")

# [Parameters]
    init_dict['ncpus'] = int(config.get('Parameters', 'ncpus').split('\t')[0].split(' ')[0])
    init_dict['ompi'] = config.get('Parameters', 'OMPI_MCA_btl_vader_single_copy_mechanism').split(
        '\t')[0].split(' ')[0]
    init_dict['USE_QSUB'] = bool(config.get('Parameters', 'USE_QSUB').split('\t')[0].split(' ')[0])
    init_dict['CompFastQ'] = bool(config.get('Parameters', 'CompFastQ').split('\t')[0].split(' ')[0]
                                  )
# [FASTQ_FILES]
    init_dict['f1'] = config.get('FASTQ_FILES', 'f1').split('\t')[0].split(' ')[0]
    init_dict['f2'] = config.get('FASTQ_FILES', 'f2').split('\t')[0].split(' ')[0]
    init_dict['f3'] = config.get('FASTQ_FILES', 'f3').split('\t')[0].split(' ')[0]
    init_dict['f4'] = config.get('FASTQ_FILES', 'f4').split('\t')[0].split(' ')[0]
# [TARGET_REGION]
    init_dict['ri'] = int(config.get('TARGET_REGION', 'ri').split('\t')[0].split(' ')[0])
# [Paths]
    init_dict['TMP_DIR'] = config.get('Paths', 'TMP_DIR').split('\t')[0].split(' ')[0]
    exe_dir = os.path.dirname(__file__)  # get directory path where script is located
    config.set('Paths', 'EXE_DIR', exe_dir)
    init_dict['EXE_DIR'] = exe_dir
    init_dict['PROJECT_DIR'] = config.get('Paths', 'PROJECT_DIR').split('\t')[0].split(' ')[0]
    config.set('Paths', 'TOOLS_DIR', f"{exe_dir}/tools")
    init_dict['TOOLS_DIR'] = f"{exe_dir}/tools"
    init_dict['INI_DIR'] = config.get('Paths', 'INI_DIR').split('\t')[0].split(' ')[0]
    init_dict['QSUB_EXE'] = config.get('Paths', 'QSUB_EXE').split('\t')[0].split(' ')[0]
# [Files]
    init_dict['CODING_SEQ'] = config.get('Files', 'CODING_SEQ').split('\t')[0].split(' ')[0]
    init_dict['AMPLICON_SEQ'] = config.get('Files', 'AMPLICON_SEQ').split('\t')[0].split(' ')[0]
    init_dict['TWIST_MUT_TABLE'] = config.get('Files', 'TWIST_MUT_TABLE').split(
        '\t')[0].split(' ')[0]
    init_dict['AA2NUC'] = config.get('Files', 'AA2NUC').split('\t')[0].split(' ')[0]
    init_dict['NUC2AA'] = config.get('Files', 'NUC2AA').split('\t')[0].split(' ')[0]
    init_dict['NUC2ABUND'] = config.get('Files', 'NUC2ABUND').split('\t')[0].split(' ')[0]
# [PRIMER_FORWARD]
    init_dict['primer_for_region1'] = config.get('PRIMER_FORWARD', 'region1').split(
        '\t')[0].split(' ')[0]
    init_dict['primer_for_region2'] = config.get('PRIMER_FORWARD', 'region2').split(
        '\t')[0].split(' ')[0]
    init_dict['primer_for_region3'] = config.get('PRIMER_FORWARD', 'region2').split(
        '\t')[0].split(' ')[0]
# [PRIMER_REVERSE]
    init_dict['primer_rev_region1'] = config.get('PRIMER_REVERSE', 'region1').split(
        '\t')[0].split(' ')[0]
    init_dict['primer_rev_region2'] = config.get('PRIMER_REVERSE', 'region2').split(
        '\t')[0].split(' ')[0]
    init_dict['primer_rev_region3'] = config.get('PRIMER_REVERSE', 'region2').split(
        '\t')[0].split(' ')[0]
# [DEPHASER]
    init_dict['DEPHASER_MAX_LENGTH'] = int(config.get('DEPHASER', 'DEPHASER_MAX_LENGTH'))
    init_dict['FORWARD_DEPHASER_SEQ'] = config.get('DEPHASER', 'FORWARD_DEPHASER_SEQ').split(
        '\t')[0].split(' ')[:3]
    init_dict['REVERSE_DEPHASER_SEQ'] = config.get('DEPHASER', 'REVERSE_DEPHASER_SEQ').split(
        '\t')[0].split(' ')[:3]
# [THRESHOLDS]
    init_dict['THRESH_MULTI_MUT'] = int(config.get('THRESHOLDS', 'THRESH_MULTI_MUT').split(
        '\t')[0].split(' ')[0])
    init_dict['FILTER'] = [int(n) for n in config.get('THRESHOLDS', 'FILTER').split(
        '\t')[0].split(' ')[:3]]

    # Writing our configuration file to 'final_init.ini'
    with open(FINAL_INIT_FN, 'w', encoding='utf-8') as configfile:
        config.write(configfile)

    return init_dict


def run_command(cmd):
    '''
    Run a system command
    '''
    print(f"\n{'*' * 30}\n{datetime.now()}: Running command: '{cmd}'")
    sps_run(cmd, shell=True, check=True)


def gzip(file_path, action="extract"):
    '''
    Compress/Uncompress files using gzip/gunzip
    '''
    if action.lower() == "compress":
        print(f"Compressing file {file_path}")
        cmd = f"gzip {file_path}"
        run_command(cmd)
    else:
        print(f"Uncompressing file {file_path}")
        cmd = f"gunzip {file_path}"
        run_command(cmd)


def main():
    '''
    Main script
    '''
    # Parse arguments
    usage = "Script to parse the reads from a long read experiment.\n" + \
            "output is a dictionary encoded as a hdf5 file"
    parser = argparse.ArgumentParser(usage, prefix_chars='-+')
    parser.add_argument('-i', '--ini', action="store", help='input initialization file')
    args = parser.parse_args()

    init_dict = {}

    if args.ini:
        try:
            if os.path.isfile(args.ini):
                init_dict = init_parse(args.ini)
        except FileNotFoundError(f"File {args.ini} was not found"):
            print("Initialization file required (-i). Type -h or --help for more help information")
            sys.exit(1)
    else:
        print("Initialization file required (-i). Type -h or --help for more help information")
        sys.exit(1)

    os.environ["OMPI_MCA_btl_vader_single_copy_mechanism"] = init_dict['ompi']
    # Print the current date and time
    print(f"Run Started: {datetime.now()}\n")

    # assign variable values
    ncpus = init_dict['ncpus']
    exe_dir = init_dict['EXE_DIR']
    region_index = init_dict['ri']
    ini_file = FINAL_INIT_FN
    direction_l = ['FORWARD', 'REVERSE']
    ba_l = ['BEFORE', 'AFTER']
    thresh_filter = ",".join([str(x) for x in init_dict['FILTER']])

    # Run ReadParser and ReadAssembler with appropriate flags
    for n, f in enumerate(fastq_files):
        fastq = init_dict[f]
        if fastq.split('.')[-1] == 'gz':
            gzip(fastq, action="extract")  # extract the fastq file if required
            fastq = ".".join(fastq.split('.')[:-1])
            init_dict[f] = fastq
        even_odd = n % 2
        if even_odd == 0:
            direction = direction_l[0]
        else:
            direction = direction_l[1]
        if n in [0, 1]:
            ba = ba_l[0]
        else:
            ba = ba_l[1]

        parser_cmd = f"""time mpirun -np {ncpus} python {exe_dir}/ReadParser.py -f {fastq} -o {
            direction}_{            ba} -e {direction_l[n % 2].lower()} -r {region_index} -c {
            ini_file} --mpi"""
        run_command(parser_cmd)

        # Compress FastQ files if defined in the .ini file
        if init_dict['CompFastQ']:
            if fastq.split('.')[-1] != 'gz':
                gzip(fastq, action="compress")  # extract the fastq file if required
                fastq = f"{fastq}.gz"
                init_dict[f] = fastq

        if even_odd != 0:
            # Combine assemblies using ReadAssembler
            assembly_out_fn = f'assembly_{ba.lower()}'
            assembly_cmd = f"""time mpirun -np {ncpus} python {exe_dir}/ReadAssembler.py --mpi -f {
                direction_l[0]}_{ba} -r {direction_l[1]}_{ba} -b -o {assembly_out_fn} -c {
                ini_file} -t {ba.lower()}"""
            run_command(assembly_cmd)

    # Combine assemblies using CombineMultiExp script
    print(f"""Will combine assemblies using [THRESHOLDS] filter = {thresh_filter
                                                                 }""")
    combiner_cmd = f"""time mpirun -np {ncpus} python {exe_dir
        }/CombineMultiExp.py -b assembly_{ba_l[0].lower()} -a assembly_{ba_l[1].lower()} -o result_thresh{
        thresh_filter.replace(',', '_')} -c {ini_file} -t {thresh_filter}"""
    run_command(combiner_cmd)
    combiner_cmd = f"""time mpirun -np {ncpus} python {exe_dir}/CombineMultiExp.py -b assembly_{
        ba_l[0].lower()} -a assembly_{ba_l[1].lower()} -o result_sum_all -c {
        ini_file} -t sum_all"""
    run_command(combiner_cmd)

    # Print the current date and time
    print("\nRun ended:", datetime.now())


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import os,re
import subprocess

def encode_barcode(s):
    """
    :param s: string of DNA
    :return: barcode ancoded as a regular expression pattern
    """
    """ 
    """
    nucD = {}
    nucD["("] = "("
    nucD[")"] = ")"
    nucD["A"] = "A"
    nucD["C"] = "C"
    nucD["G"] = "G"
    nucD["T"] = "T"
    nucD["U"] = "U"
    nucD["R"] = "AG"
    nucD["Y"] = "CT"
    nucD["S"] = "GC"
    nucD["W"] = "AT"
    nucD["K"] = "GT"
    nucD["M"] = "AC"
    nucD["B"] = "CGT"
    nucD["D"] = "AGT"
    nucD["H"] = "ACT"
    nucD["V"] = "ACG"
    nucD["N"] = "ACGT"

    pattern = ""
    for n in s:
        if len(nucD[n]) == 1:
            pattern += nucD[n]
        else:
            pattern += "[{}]".format(nucD[n])
    return pattern

def get_dicaa():
    d3to1 = {'ALA':'A','CYS':'C','ASP':'D','GLU':'E','PHE':'F','GLY':'G',\
             'HIS':'H','ILE':'I','LYS':'K','LEU':'L','MET':'M','ASN':'N',\
             'PRO':'P','GLN':'Q','ARG':'R','SER':'S','THR':'T','VAL':'V',\
             'TRP':'W','TYR':'Y','TER':'Z','XXX':'X'}
    d1to3 = {}
    for key,val in d3to1.items():
        d1to3[val] = key
    return d3to1, d1to3

def get_reverse(s):
    """

    :param s: anysequence
    :return: the reversed order sequence
    """
    reversed_sequence = ""
    x = s[::-1]
    for c in x:
        if c=="(":
            reversed_sequence += ")"
        elif c==")":
            reversed_sequence += "("
        else:
            reversed_sequence += c
    return reversed_sequence


def get_complement(s):
    """
    :param s: string of DNA
    :return: complement
    """
    """ Returns complementary DNA sequence even with ambiguous nucleotides.
    """
    nucD = {}
    nucD["\n"] = ""
    nucD["("] = "(" # useful for regular expression tracking
    nucD[")"] = ")" # useful for regular expression tracking
    nucD["A"] = "T"
    nucD["C"] = "G"
    nucD["G"] = "C"
    nucD["T"] = "A"
    nucD["U"] = "A"
    nucD["R"] = "Y"
    nucD["Y"] = "R"
    nucD["S"] = "S"
    nucD["W"] = "W"
    nucD["K"] = "M"
    nucD["M"] = "K"
    nucD["B"] = "V"
    nucD["D"] = "H"
    nucD["H"] = "D"
    nucD["V"] = "B"
    nucD["N"] = "N"
    complement = ""
    for n in s:
        complement += nucD[n]
    return complement

def count_file_lines(input_file):
    """

    :param input_file:
    :return: integer with number of lines
    """
    return int(subprocess.check_output(["wc","-l","{}".format(input_file)]).split()[0])

def convert_list_bytes2string(l):
    x = []
    for e in l:
        if isinstance(e,bytes):
            try:
                e = e.decode()
            except:
                e = ' '
        elif isinstance(e, list):
            e = convert_list_bytes2string(e)
        x.append(e)
    return x

def convert_dic_bytes2string(d):
    """
    Python Code to convert ByteString key:value
        pair of dictionary to String.
    :param d: dictionnary of bytesString (often got from unpickling python2 dict)
    :return: dictionnary with String format in key:values
    """
    x = {}
    for key, value in d.items():
        if isinstance(key,bytes):
            key = key.decode()
        if isinstance(value,bytes):
            value = value.decode()
        elif isinstance(value, dict):  # recursivity if dict of dict
            value = convert_dic_bytes2string(value)
        elif isinstance(value,list): # recursivity if list
            value = convert_list_bytes2string(value)
        x[key] = value
    return x

def get_seq_from_fasta(fasta_file):
    """

    :param file:
    :return:
    """
    fin = open(fasta_file).readlines()
    seq = ""
    for l in fin:
        if l[0] == ">":
            header = l[1:].strip()
        else:
            if len(l.split()) > 0:
                seq += l.strip("\n")
    return seq

def get_mutation_in_read(read, coding_seq, index,  thresh_multi_mut=10):
    """

    :param read:
    :param coding_seq:
    :param index:
    :return:
    """
    len_coding_seq = len(coding_seq)
    mutation_code = ""
    if index + len(read) <= len(coding_seq):
        len_read_limit = len(read)
    else:
        len_read_limit = len(coding_seq) - index
    if read[:len_read_limit] == coding_seq[index:index+len(read[:len_read_limit])]:
        return "W"

    separator = ''
    FLAG_MUTANT = 0
    for ii,n in enumerate(read):
        index_in_coding_seq = index + ii
        if index_in_coding_seq >= len_coding_seq:
            break
        elif n == coding_seq[index_in_coding_seq]:
            pass
        else:
            FLAG_MUTANT += 1
            mutation_code += separator
            mutation_code += str(index_in_coding_seq) + n
            if separator == '_':
                FLAG_MUTANT += 1
            if index_in_coding_seq%3 != 2:
                separator = '.'

        if FLAG_MUTANT and index_in_coding_seq%3 == 2:
            separator = '_'
        if FLAG_MUTANT > thresh_multi_mut:
            return "T"

    return mutation_code

def merge_dict_of_lists(dol1, dol2):
    """
    Useful to combine multiple dictionaries generated after mpi instructions
    :param dol1: dict 1 of lists
    :param dol2:  dict 2 of lists
    :return: a merge with lists extended
    """
    keys = set(dol1).union(dol2)
    no = []
    return dict((k, dol1.get(k, no) + dol2.get(k, no)) for k in keys)

def data_merge(a, b,sum_values = False):
    """merges b into a and return merged result
    By default the new numerical values replace the previous
    setting sum_values = True triggers the addition operation between values targetted by the same key

    NOTE: tuples and arbitrary objects are not handled as it is totally ambiguous what should happen"""
    key = None
    if a is None or isinstance(a, (str, int, float)):
        # border case for first run or if a is a primitive
        if sum_values and isinstance(a, (str, int, float)):
            a += b
        else:
            a = b
    elif isinstance(a, list):
        # lists can be only appended
        if isinstance(b, list):
            # merge lists
            a.extend(b)
        else:
            # append to list
            a.append(b)
    elif isinstance(a, set):
        if isinstance(b, set):
            # merge sets
            a.update(b)
        else:
            # append to set
            a.add(b)

    elif isinstance(a, dict):
        # dicts must be merged
        if isinstance(b, dict):
            for key in b:
                if key in a:
                    a[key] = data_merge(a[key], b[key], sum_values = sum_values)
                else:
                    a[key] = b[key]
    return a

def get_aamut(dn2a, coding_seq, mutation_code, aa1letter = True):
    """
    Convert into aminoacids code the nucleotide code of the mutations
    :param dn2a:
    :param coding_seq:
    :param mutation_code:
    :return: list_aamut, list_nucmut
    """
    codons = mutation_code.split('_')
    list_aamut = []
    list_nucmut = []
    if aa1letter:
        d3to1, d1to3 = get_dicaa()
    for codon in codons:
        nmuts = codon.split('.')
        try:
            mut_index = [int(x[:-1]) for x in nmuts] # as [393,394]
        except ValueError:
            #print(mutation_code)
            return None,None
        mut_nuc = [x[-1] for x in nmuts] # as ['G','T']
        dmut = dict(zip(mut_index,mut_nuc))
        start_codon = 3*(mut_index[0]//3)
        wt_codon = coding_seq[start_codon:start_codon+3]
        codon_index = range(start_codon,start_codon+3)
        mut_codon = ""
        for ii,n in enumerate(codon_index):
            if n in dmut:
                mut_codon += dmut[n]
            else:
                mut_codon += wt_codon[ii]
        aa_wt = dn2a[wt_codon]
        if aa1letter:
            aa_wt = d3to1[aa_wt]
        try:
            aa_mut = dn2a[mut_codon]
        except:
            aa_mut = "XXX"
        if aa1letter:
            aa_mut = d3to1[aa_mut]
        aa_mutname = aa_wt + str(mut_index[0]//3+1) + aa_mut
        nuc_mutname = wt_codon + str(start_codon) + mut_codon
        list_aamut.append(aa_mutname)
        list_nucmut.append(nuc_mutname)
    return list_aamut, list_nucmut

def jaccard_dist_equallen(bc1,bc2):
    """
    Compute the number of different nucleotides between the bases of two barcodes of equal length
    :param bc1: sequence of the first barcode ...N1N2N3...
    :param bc2: sequence of the second barcode ...N'1N'2N'3...
    :return: jaccard distance between the two barcodes
    """

    # Once we checked the dephasing tails are identical, we focus on the UMIs
    jdistance = 0
    for cursor in range(len(bc1)):
        if bc1[cursor] != bc2[cursor]:
            jdistance += 1
    return jdistance


def jaccard_dist(bc1,bc2,n_flank=7,check_dephaser=True):
    """
    Compute the number of different nucleotides between the bases of two barcode
    It analyses the n_flank nucleotides surrounding the X central between 2 UMI
    Can check the dephaser nucleotide in 5' and 3' in n_flank-1 and n_flank+1.
    If check_dephaser is True, it will return a distance n_flank*2
    :param bc1: sequence of the first barcode ...N4N3N2N1XN1N2N3...
    :param bc2: sequence of the second barcode ...N'4N'3N'2N'1XN'1N'2N'3...
    :return:
    """
    index_X1 = bc1.index('X')
    index_X2 = bc2.index('X')
    if check_dephaser:
        # Small tails at the extremities of the reads to dephase
        i5_dephaser1 = index_X1 - n_flank - 1
        i5_dephaser2 = index_X2 - n_flank - 1
        i3_dephaser1 = index_X1 + n_flank + 1
        i3_dephaser2 = index_X2 + n_flank + 1
        # If they are difference reads are necessarily belonging to different categories.
        if bc1[i5_dephaser1] != bc2[i5_dephaser2] or bc1[i3_dephaser1] != bc2[i3_dephaser2]:
            jdistance = 2*n_flank
            return jdistance
    # Once we checked the dephasing tails are identical, we focus on the UMIs
    jdistance = 0
    for cursor in range(n_flank):
        bc1_index = index_X1 + cursor + 1
        bc2_index = index_X2 + cursor + 1
        if bc1[bc1_index] != bc2[bc2_index]:
            jdistance += 1
        bc1_index = index_X1 - cursor - 1
        bc2_index = index_X2 - cursor - 1
        if bc1[bc1_index] != bc2[bc2_index]:
            jdistance += 1
    return jdistance

def compute_minimal_id_with_bcref(bc,set_bc,exclude_bc = False):
    """
    Analyze the minimal distance between a barcode and a set of other barcodes for which the barcode may originate.
    The idea is that noise may have been responsible for the generation of that barcode
    :param bc: a specific barcode
    :param set_bc: a set of barcodes which might be at the origin of bc.
    :param exclude_bc: Do not count when bc and bcref are identical
    :return: minimal jaccard-like distance
    """
    min_diff = 14
    #cc = 0
    for bcref in set_bc:
        #cc+=1
        if exclude_bc:
            if bcref == bc:
                continue
        jdist = jaccard_dist(bc,bcref)
        #if cc==1 and jdist !=14:
        #    print(jdist,bc,bcref)
        if jdist < min_diff:
            min_diff = jdist
    return min_diff

def parse_twist_devis_table(twist_file, dn2a, d3to1, start=None, stop=None, info_codon_index=None):
    """
    :param twist_file: csv table generated for the order at Twist
    :param dn2a: dic extracted from the nuc2aa.pickle
    :param d3to1: dic to translate 3-letter code aa into 1
    :param start: residue index of the beginning of the sequence in the csv file
    :param stop: residue index of the end of the sequence in the csv file
    :param info_codon_index:
    :return: dictionnary integrating all the information about the mutation landscape
            dic_mutnuc[mut_nuc_name] mut_nuc_name as "125G.126T.127T" or "125C.127A"
            dic_mutnuc[mut_nuc_name]['nuc_wt'] -> wt_nucleotide
            dic_mutnuc[mut_nuc_name]['AA_index'] -> resi
            dic_mutnuc[mut_nuc_name]['AA_wt'] -> wt_aminoacid
            dic_mutnuc[mut_nuc_name]['nuc_mut'] -> mutated_nucleotide
            dic_mutnuc[mut_nuc_name]['AA_mut'] -> mutated amino acid
            dic_mutnuc[mut_nuc_name]['nuc_syn'] -> Synonymous mutations (set of other mut_nuc_name)
             dic_mutnuc[mut_nuc_name]['code_type'] -> Either '30'->WT AA with 3 nuc mutations,
                                                or      '20'->WT AA with 2 nuc mutations,
                                                or      '10'->WT AA with 1 nuc mutations,
                                                or mutated AA with '3', '2' or '1' nucleotide
           * Historically full information is given by:
            dic_mutnuc[mut_nuc_name]['full_mut_AA_name'] -> 'nuc_mutation@@aa_mutation'
            * Shorter information (no wt nuc) is given by:
            dic_mutnuc[mut_nuc_name]['short_mut_AA_name'] -> 'nuc_mutation@@aa_mutation'

    Variables defined in the script but not returned
        list_aai -> Asf1 absolute index in structure
        dic_mut[<Asf1 absolute index in structure>]['i2d'] -> <Asf1 index from 0 to n>
        dic_ftwist[<first column of the twist file=legend or codon>] -> line following the 1st col
    """
    ftwist = open(twist_file).readlines()
    #print("> Reading the mutation profile table : ",twist_file)
    dic_mut = {}
    dic_ftwist = {}

    dic_mutnuc = {}
    set_mutnuc = set()
    list_mut_nuc_name = []

    if start:
        START = start
    if stop:
        STOP = stop
    if info_codon_index:
        CODON_INDEX_INFO_LINE = info_codon_index

    if not start:
        START = 1
    #print("> The first index of the AA considered in mutation table is :", START)

    set_possible_codons = []  # set of list[codon,AA_3letter]
    FOUND_AAindex = False
    for l in ftwist:
        s = [x.strip() for x in l.split(';')]
        if s[0] == '':
            continue
        dic_ftwist[s[0]] = s[1:]
        if len(s[0]) == 3:
            set_possible_codons.append([s[0], s[1]])
        if s[0] == 'AA_index' or s[0] == 'PROTEIN_index':
            FOUND_AAindex = True
            list_aai = s[2:] # List of indexes matching coding sequence
            for ii in range(1, len(list_aai) + 1):
                dic_mut[list_aai[ii - 1]] = {}
                # matches REAL resi index to index d which is used to retrieve the information in the other columns
                dic_mut[list_aai[ii - 1]]['i2d'] = ii
    if not FOUND_AAindex:
        print("> The mutation profile table file is lacking the AA_index line. Check the correct format !")
    if not stop:
        STOP = int(list_aai[ii - 1])

    #print("> The last index of the AA considered in mutation table is :",STOP)

    # Definition for IP scanned region
    for resi in range(START, STOP + 1):
        ini_nuc = dic_ftwist['NUC_wt'][dic_mut[str(resi)]['i2d']]
        # ini_nuc is wt codon of every mutated position resi
        try:
            codon_index = int(dic_ftwist[CODON_INDEX_INFO_LINE][dic_mut[str(resi)]['i2d']])
        except:
            codon_index = (resi-1)*3

        n = 0
        for key in dic_ftwist:
            # key is <first column of the twist file>
            if len(key) != 3:
                continue
            """
            We read only the codons
            """
            # for every resi we get the equivalent column index and get whether codon is mutated or not
            mut_nuc = dic_ftwist[key][dic_mut[str(resi)]['i2d']]
            if mut_nuc == '':
                # codon not mutated
                continue
            n += 1
            list_mutnuc = []
            list_mutnuc_withwt = []
            #ini_nuc is wt codon
            for jj, nn in enumerate(ini_nuc):
                if nn != mut_nuc[jj]:
                    nuc_index = codon_index + jj
                    list_mutnuc.append('{}{}'.format(nuc_index, mut_nuc[jj]))
                    list_mutnuc_withwt.append('{}{}{}'.format(nn, nuc_index, mut_nuc[jj]))
            if len(list_mutnuc) > 0:
                mut_nuc_name = '.'.join(list_mutnuc)
                set_mutnuc.add(mut_nuc_name)
                dic_mutnuc[mut_nuc_name] = {}
                list_mut_nuc_name.append(mut_nuc_name)
                dic_mutnuc[mut_nuc_name]['nuc_wt'] = ini_nuc
                dic_mutnuc[mut_nuc_name]['AA_index'] = resi
                dic_mutnuc[mut_nuc_name]['AA_wt'] = dic_ftwist['AA_wt'][dic_mut[str(resi)]['i2d']]
                dic_mutnuc[mut_nuc_name]['nuc_mut'] = mut_nuc
                dic_mutnuc[mut_nuc_name]['AA_mut'] = dn2a[mut_nuc]
                dic_mutnuc[mut_nuc_name]['type'] = dn2a[mut_nuc]
                dic_mutnuc[mut_nuc_name]['nuc_syn'] = None # initialized here, filled below
                label1 = '.'.join(list_mutnuc_withwt)
                label2 = '{}{}{}'.format(d3to1[dic_mutnuc[mut_nuc_name]['AA_wt']], resi,
                                         d3to1[dic_mutnuc[mut_nuc_name]['AA_mut']])
                if dic_mutnuc[mut_nuc_name]['AA_wt'] == dic_mutnuc[mut_nuc_name]['AA_mut']:
                    dic_mutnuc[mut_nuc_name]['code_type'] = '{}0'.format(len(list_mutnuc))
                else:
                    dic_mutnuc[mut_nuc_name]['code_type'] = '{}'.format(len(list_mutnuc))

                dic_mutnuc[mut_nuc_name]['full_mut_AA_name'] = '@@'.join([label1, label2])
                dic_mutnuc[mut_nuc_name]['short_mut_AA_name'] = '@@'.join([mut_nuc_name,label2])

            else:
                print('problem with {},{}'.format(resi, key))
    #print('> The number of mutants in the sequenced region should be : {}'.format(len(set_mutnuc)))
    dic_mutnuc['order'] = list_mut_nuc_name[:]
    # Filling the synonymous mutants
    for mut1 in dic_mutnuc['order']:
        for mut2 in dic_mutnuc['order']:
            if mut1 == mut2:
                continue
            if dic_mutnuc[mut1]['AA_index'] == dic_mutnuc[mut2]['AA_index'] \
                    and dic_mutnuc[mut1]['AA_mut'] == dic_mutnuc[mut2]['AA_mut']:
                if dic_mutnuc[mut1]['code_type'][-1] == '0': # We don't consider WT synonymous a priori
                    continue
                if not dic_mutnuc[mut1]['nuc_syn']:
                    dic_mutnuc[mut1]['nuc_syn'] = set()
                if not dic_mutnuc[mut2]['nuc_syn']:
                    dic_mutnuc[mut2]['nuc_syn'] = set()
                dic_mutnuc[mut1]['nuc_syn'].add(mut2)
                dic_mutnuc[mut2]['nuc_syn'].add(mut1)
    return dic_mutnuc

def get_quality_score(quality_sequence):
    """

    :param quality_sequence:
    :return:
    """
    score = 1.
    for ascii_char in quality_sequence:
        if ord(ascii_char)-33 < 42 and ord(ascii_char)-33 > 0:
            score *= 1-10**(-(ord(ascii_char)-33)/10.)
    return score

def get_mutation_type(mutant, ref_mut_db=None):
    """

    :param mutant:
    :param ref_mut_db:
    :return: if canonical mut returns ["MONO"],["DI"],["TRI"]
             if single mutaa non canonical returns   ["BAD","MONO"],["BAD","DI"],["BAD","TRI"]
             if multiple mutaa returns ["BAD",<int nbmutaa>,<type of the largest nbmutnuc either "MONO","DI"or"TRI">]
    """
    mutation_type = []
    dtypecode = dict(zip(range(1,4),["MONO","DI","TRI"]))
    if ref_mut_db:
        if not mutant in ref_mut_db:
            mutation_type.append('BAD')
    mutaa = mutant.split('_')
    if len(mutaa) > 1:
        max_nbnucmut = max([len(mut.split('.')) for mut in mutaa])
        mutation_type.append("{}".format(len(mutaa)))
        mutation_type.append(dtypecode[max_nbnucmut])
    else:
        nbnucmut = len(mutant.split('.'))
        mutation_type.append(dtypecode[nbnucmut])
    return mutation_type



#!/usr/bin/env python3

import lr_tools as lt

import sys
import os
import re

from datetime import datetime
import time
import tempfile
import argparse
import pickle
import fileinput
import pprint
from copy import deepcopy
from mpi4py import MPI
from itertools import islice

import fileinput as FI
import h5py
import warnings

warnings.filterwarnings("ignore")


try:
    import configparser
except:
    import ConfigParser as configparser
from configparser import ExtendedInterpolation



class Parser:
    def __init__(self, fastq,
                 exp_orientation,
                 region,
                 config_file,
                 use_mpi,
                 outputfile,
                 verbose = False, cleanup = True):
        self.fastq = fastq
        self.use_mpi = use_mpi
        self.exp_orientation = exp_orientation
        self.region = int(region)
        self.config_file = config_file
        self.verbose = verbose
        self.cleanup = cleanup

        self.d3to1, self.d1to3 = lt.get_dicaa()

        # Parameter definition
        self.thresh_quality_reads = 0.3
        self.use_relax_pattern_in_constant_region = False

        self.start_mpi()
        if self.rank == 0:
            print("\n>>> Counting number of lines.")
        self.number_of_lines = self.get_nlines()

        if self.rank == 0:
            print("\n>>> Reading the configuration file.")
        self.read_config_file()

        if self.rank == 0:
            print("\n>>> Mapping the index of the primers on the reference sequence.")
        self.map_indexes_inrefseq()

        if self.rank == 0:
            if self.exp_orientation == 'forward':
                print("> FORWARD PATTERN_UMI:", self.forward_umi)
                print("> FORWARD CONSTANT_PRIMER:",self.forward_cst)
            if self.exp_orientation == 'reverse':
                print("> REVERSE CONSTANT_PRIMER:",self.reverse_cst)
                print("> REVERSE PATTERN_UMI:",self.reverse_umi)

        if self.rank == 0:
            print("\n>>> Extracting the expected mutation patterns.")
        self.get_dic_expect_mutations()

        if self.rank == 0:
            print("\n>>> Parsing the fastq with {} cpus".format(self.size))
        self.split_input(outputfile, rank=self.rank, size=self.size)


    def start_mpi(self):
        if self.use_mpi:
            # MPI communication
            self.comm = MPI.COMM_WORLD
            self.rank = self.comm.Get_rank()
            self.size = self.comm.Get_size()
        else:
            self.comm = None
            self.rank = 0
            self.size = 1

    def get_nlines(self):
        if self.rank == 0:
            number_of_lines = lt.count_file_lines(self.fastq)
            #print("I finished")
        else:
            number_of_lines = None
        if self.use_mpi:
            self.comm.Barrier()
            number_of_lines = self.comm.bcast(number_of_lines, root=0)
        if self.rank==0:
            print("> The number of reads in the input file is: {}\n".format(number_of_lines//4))
        return number_of_lines

    def get_dic_expect_mutations(self):
        """

        :return:
        """
        self.dmut = lt.parse_twist_devis_table(self.mutfile, self.dn2a, self.d3to1, start=None, stop=None, info_codon_index=None)
        #if self.verbose:
        #   pprint.pprint(self.dmut)

    def get_position_primer(self, amplicon_seq, constant_primer, orient="forward"):
        """
        Method which reduces the primer until it matches the amplicon_seq
        :param amplicon_seq: The full amplicon sequence with its flanking regions
        :param constant_primer: The primer of interest matching parly to the amplicon sequence
        :return: The delimitations for the region of the primer which matched
        """
        primer = constant_primer
        #print(primer,amplicon_seq)
        for ii in range(len(constant_primer)):
            pat = re.compile(primer)
            match = pat.search(amplicon_seq)
            if not match:
                if orient=="forward":
                    primer = constant_primer[ii+1:]
                elif orient=="reverse":
                    primer = constant_primer[:-ii-1]
            else:
                break

        if not match:
            return None, None, None
        else:
            limits = match.span()
            if self.rank == 0:
                print ("> The constant part of the {} primer '{}' matches over the sequence '{}' in positions {}".format(
                    orient,constant_primer,primer,limits))
            return limits[0], limits[1], primer

    def map_indexes_inrefseq(self):
        """
        From the data extracted in the .ini files, this method automatically
        extracts the index where the analysis of the variants should start and stop
        :return: indexes for xxx
        """

        mcodseq2ampli = re.search(self.coding_seq, self.amplic_seq)
        try:
            self.start_codseq_in_ampliseq, self.stop_codseq_in_ampliseq = mcodseq2ampli.span()
        except AttributeError:
            print("!!! ERROR: The coding sequence and amplicon file were not properly implemented.\n"
                  "-> Amplicon should correspond to the largest DNA fragment embedding the sequences from the vector used as constant parts, etc... \n"
                  "-> Coding sequence should be entirely embedded in the amplicon sequence\n")
            sys.exit()

        start_f, stop_f, matched_f = self.get_position_primer(self.amplic_seq, self.forward_cst,orient='forward')
        start_r, stop_r, matched_r = self.get_position_primer(self.amplic_seq, self.reverse_cst,orient='reverse')
        if start_f is None:
            print("!!! ERROR: The forward primer could not be found in the amplicon. Check your inputs")
            sys.exit()
        if start_r is None:
            print("!!! ERROR: The reverse primer could not be found in the amplicon. Check your inputs")
            sys.exit()

        self.start_read_in_codseq = stop_f - self.start_codseq_in_ampliseq
        self.stop_read_in_codseq = start_r - self.start_codseq_in_ampliseq
        if self.rank==0:
            print("> Positions of Coding Sequence Matching the Amplicon Seq. : {} - {} ".format(
                self.start_codseq_in_ampliseq, self.stop_codseq_in_ampliseq))
            print("> Forward Primer in Coding Sequence Stops at : {} ".format(self.start_read_in_codseq))
            print("> Reverse Primer Starts in Coding Sequence Starts at : {} ".format(self.stop_read_in_codseq))

    def split_input(self, outputfile, rank=0, size=1):
        """Split an input with Nlines so that every rank creates its fraction and works on it
        """

        if self.exp_orientation == 'forward':
            choice_dephaser = "({})".format("|".join(self.forward_dephaser))
            pattern1 = r"{dephaser}([ATCG]{{{len_umi}}})({cst_anchor})([ATGC]+)".format(
                dephaser=choice_dephaser,
                len_umi=len(self.forward_umi) - 2,
                cst_anchor=self.forward_cst) # we remove - 2 because of barckets
            # This is a test to increase the number of reads analyzed.
            # With pattern1_relaxed, we consider only the size of the constant region and not the sequence itself.
            pattern1_relaxed = r"^{dephaser}([ATCG]{{{len_umi}}})([ATCG]{{{len_cst_anchor}}})([ATGC]+)".format(
                dephaser=choice_dephaser,
                len_umi=len(self.forward_umi) - 2,
                len_cst_anchor=len(self.forward_cst)) # we remove - 2 because of barckets
        elif self.exp_orientation == 'reverse':
            choice_dephaser = "({})".format("|".join(self.reverse_dephaser))
            pattern1 = r"([ATGC]+)({cst_anchor})([ATCG]{{{len_umi}}}){dephaser}".format(
                dephaser=choice_dephaser,
                len_umi=len(self.reverse_umi) - 2,
                cst_anchor=self.reverse_cst) # we remove - 2 because of barckets
            # This is a test to increase the number of reads analyzed.
            # With pattern1_relaxed, we consider only the size of the constant region and not the sequence itself.
            pattern1_relaxed = r"([ATGC]+)([ATCG]{{{len_cst_anchor}}})([ATCG]{{{len_umi}}}){dephaser}$".format(
                dephaser=choice_dephaser,
                len_umi=len(self.reverse_umi) - 2,
                len_cst_anchor=len(self.reverse_cst))  # we remove - 2 because of barckets
        #print(pattern1)
        if rank == 0:
            print("> Splitting the input file into {} parts".format(size))

        self.outputfile = open(outputfile + ".rank{}".format(rank), 'w')
         # Ditributing the work for every rank
        full_input = self.fastq
        size_chunks = (int(self.number_of_lines) / 4) / self.size + 1 # we want to recover blocks of 4 lines
        index_start = int(rank * (size_chunks * 4))
        if rank == size - 1:
            index_stop = None
        else:
            index_stop = int((rank + 1) * (size_chunks * 4))
        if self.verbose:
            print("!! rank {}, will parse lines from {} to {}.".format(rank, index_start, index_stop))

        hit = 0
        read_index = 0
        rank_index = 10 + rank
        if self.rank == 0:
            count = 0
        else:
            count = None
        count = self.comm.scatter([count for i in range(self.size)],root=0)
        with open(full_input) as file_obj:
            n = 1
            object_to_iterate = islice(file_obj, index_start, index_stop)
            # for line in islice(file_obj, index_start, index_stop):
            for line in object_to_iterate:
                # Parse the fastq file
                l = line[: -1]
                if l[0] == "@":
                    if self.verbose:
                        print("\n############ NEW_READ #############\n")
                    n=1
                    FLAG_REACH_END_READ = False
                    read_index += 1
                    rdint = int('%d%09d' % (rank_index, read_index))
                    rdhex = hex(rdint)
                    #print(rdhex)
                    header = l
                    n += 1
                    KEEP_READ = False
                    PRIMER_DEFECT = False
                    BARCODE_DEFECT = False
                elif n == 4:
                    quality = line.strip()
                    quality_score = lt.get_quality_score(quality)
                    if self.verbose:
                        print("quality_score = {:.3f}".format(quality_score))
                    n += 1
                elif n == 3:
                    n += 1
                elif n == 2:
                    n += 1
                    # extract umi
                    if self.exp_orientation=='forward':
                        seq = line.strip()
                    elif self.exp_orientation=='reverse':
                        seq = lt.get_reverse(lt.get_complement(l.strip()))

                if n == 5:
                    count +=1
                    match = re.search(pattern1, seq)
                    if match:  # primer is detected
                        KEEP_READ = True
                    else:
                        # Try to read the reads even in case of primer defects
                        # Test the cases in which the constant part would have been mutated
                        # Only the length of the constant stretch is analyzed
                        if self.use_relax_pattern_in_constant_region and quality_score >= self.thresh_quality_reads:
                            match = re.search(pattern1_relaxed, seq)
                            if match:
                                KEEP_READ = True
                            else:
                                KEEP_READ = False
                                PRIMER_DEFECT = True
                        else:
                            KEEP_READ = False
                            PRIMER_DEFECT = True
                    if match:
                        if self.exp_orientation == 'forward':
                            dephaser = match.group(1)
                            start_seq, stop_seq = match.span(4)
                            umi = dephaser + match.group(2)
                        if self.exp_orientation == 'reverse':
                            dephaser = match.group(4)
                            start_seq, stop_seq = match.span(1)
                            umi = match.group(3) + dephaser

                        if self.verbose:
                            print("First filter for constant primer detection passed")
                            print(start_seq, stop_seq,
                                                  '\n',seq[start_seq:start_seq+30],'...',seq[stop_seq-30:stop_seq])
                            print(self.start_read_in_codseq,self.start_read_in_codseq%3,self.start_read_in_codseq//3)
                            print("len_coding_seq:",len(self.coding_seq))
                            print("coding_seq:",self.coding_seq)
                            print("####\n")
                            print(self.coding_seq[self.start_read_in_codseq:self.start_read_in_codseq+30])
                            print(seq[:30])
                    #
                    # Get the mutation code in nucleotides
                    #
                    if KEEP_READ:
                        if self.exp_orientation == 'forward':
                            mcode = lt.get_mutation_in_read(seq[start_seq:],
                                                        self.coding_seq,
                                                        self.start_read_in_codseq,
                                                        thresh_multi_mut = self.thresh_multi_mut)
                        if self.exp_orientation == 'reverse':
                            #print("start_index = {}".format(self.stop_read_in_codseq-len(seq[:stop_seq])))
                            mcode = lt.get_mutation_in_read(seq[:stop_seq],
                                                        self.coding_seq,
                                                        self.stop_read_in_codseq-len(seq[:stop_seq]),
                                                        thresh_multi_mut = self.thresh_multi_mut)
                        if self.verbose:
                            print("## Keep:{}".format(header))
                            print("UMI:",umi)
                            print("SEQ:",seq[start_seq:start_seq+30])
                            print("mcode:",mcode)
                            # getaa will raise an error in nucleotide is N
                            print(lt.get_aamut(self.dn2a, self.coding_seq.strip(), mcode))
                            print("\n")
                    else:
                        if self.verbose:
                            print("## Discard:{}\n".format(header))
                    #

                    if KEEP_READ:
                        if mcode in self.dmut: # for instance compatible with twist or trinex mutational pattern
                            mcode = "O\t" + mcode
                        else:
                            mcode = "N\t" + mcode
                        self.outputfile.write("{}\t{:.2f}\t{}\t{}\n".format(rdhex, quality_score, mcode, umi))
                    else:
                        if PRIMER_DEFECT:
                            self.outputfile.write("{}\t{:.2f}\tD\tP\n".format(rdhex, quality_score))
                            #self.outputfile.write(seq+"\n")
                            #self.outputfile.write(quality+"\n")
                        elif BARCODE_DEFECT:
                            self.outputfile.write("{}\t{:.2f}\tD\tB\n".format(rdhex, quality_score))
                        else:
                            # Normally there should not be D D but this can be checked
                            self.outputfile.write("{}\t{:.2f}\tD\tD\n".format(rdhex, quality_score))

        if rank == 0:
            print(
                "Start of the mutant analysis at position in the coding sequence:{}".format(self.start_read_in_codseq))
            print("First codon analyzed indexed as aminoacid: {}".format(self.start_read_in_codseq // 3 + 1))
        print("RANK:{} -> {}".format(self.rank,count))
        if self.use_mpi:
            count = self.comm.gather(count,root=0)
            if self.rank == 0:
                sum_count = sum(count)
        else:
            sum_count = 0
        self.outputfile.close()
        if rank == 0:
            print("SUM_RANK: {}".format(sum_count))
            listfiles = []
            for r in range(size):
                listfiles.append(outputfile + ".rank{}".format(r))
            with open(outputfile,"w") as fout, fileinput.input(listfiles) as fin:
                for line in fin:
                    fout.write(line)
            for fname in listfiles:
                os.remove(fname)


    def read_config_file(self):
        """
        Recovers the parameters required to process the input files
        region: int (1,2,3,... or N) defining which region of the gene was sequenced
        All variable are defined
        """
        config = configparser.ConfigParser(interpolation=ExtendedInterpolation())
        config.read(self.config_file)
        if self.rank==0:
            print(" Path to the config file: {}\n Content: \n".format(self.config_file))
            print(open(self.config_file).read())
            print("                 ****  END CONFIG FILE ****                 ")

        self.forward_primer = config.get("PRIMER_FORWARD","region{}".format(self.region))
        self.reverse_primer = config.get("PRIMER_REVERSE","region{}".format(self.region))

        #
        # Parameters for barcode retrieving
        #
        # forward should such as "(HHNHHNH)ATTCCGGGTCTTCCATGGCG" ; reverse as CCGACCT(DNDDNDD)

        pat_umi_cst = re.compile(r"(\(\w*\))?([ATGC]+)(\(\w*\))?")
        try:
            f_umi_cst = pat_umi_cst.findall(self.forward_primer)
            r_cst_umi = pat_umi_cst.findall(self.reverse_primer)
            self.forward_cst = f_umi_cst[0][1]
            self.forward_umi = f_umi_cst[0][0]
            self.reverse_cst = r_cst_umi[0][1]
            self.reverse_umi = r_cst_umi[0][2]
        except IndexError:
            print("ERROR: Could not parse the primers. Please check primers format in .ini file.")
            sys.exit()
        #
        # Dephaser are short nucleotide stretches that can be added to the end of the amplicons.
        #
        ndephaser          = int(config.get("DEPHASER", "DEPHASER_MAX_LENGTH"))
        if ndephaser > 0:
            self.forward_dephaser = config.get("DEPHASER", "FORWARD_DEPHASER_SEQ").split()
            self.reverse_dephaser = config.get("DEPHASER", "REVERSE_DEPHASER_SEQ").split()
        else:
            self.forward_dephaser = None
            self.reverse_dephaser = None

        self.coding_seq    = lt.get_seq_from_fasta(config.get("Files", "CODING_SEQ"))
        self.amplic_seq    = lt.get_seq_from_fasta(config.get("Files", "AMPLICON_SEQ"))

        self.mutfile       = config.get("Files", "TWIST_MUT_TABLE")

        faa2nuc            = config.get("Files", "AA2NUC")
        self.da2n = pickle.load(open(faa2nuc,'rb'),encoding='bytes') # due to py2 to 3
        self.da2n = lt.convert_dic_bytes2string(self.da2n) # due to py2 to 3
        fnuc2aa   = config.get("Files", "NUC2AA")
        self.dn2a = pickle.load(open(fnuc2aa,'rb'))#,encoding='bytes')

        #nuc2abundance = config.get("Files", "NUC2ABUND")
        self.thresh_multi_mut = int(config.get("THRESHOLDS", "THRESH_MULTI_MUT"))








def Main():

    try:
        default_configfile = os.path.join(os.path.dirname(os.path.abspath(os.readlink(os.path.abspath(__file__)))),
                                  "NGS_DMS.ini")
    except:
        default_configfile = os.path.join(os.path.dirname(os.path.abspath(os.path.realpath(os.path.abspath(__file__)))),
                                  "NGS_DMS.ini")


    usage = "Script to parse the reads from a long read experiment.\n" + \
            "output is a dictionary encoded as a hdf5 file"
    parser = argparse.ArgumentParser(usage)
    parser.add_argument('-f', '--fastq', action="store", help='input fastq file')
    parser.add_argument('-o', '--outputfile', action="store", help='output mutant analysis file')
    parser.add_argument('-e', '--exp_orientation', choices = ['forward', 'reverse'], action="store", help='experiment is reverse or forward')
    parser.add_argument('-r', '--region_index', action="store", help='Which region of the protein targeted. Identified by an integer 1 to N.')
    parser.add_argument('-c', '--config_file', default=default_configfile, action="store", help='name of the configuration file')
    parser.add_argument('--mpi', action="store_true", default=False, help='triggers mpi running')
    parser.add_argument('--verbose', action="store_true", default=False, help="print extra info")
    parser.add_argument('--cleanup', action="store_true", default=False, help="cleanup temporary files after execution")

    if len(sys.argv) == 1:
        print("type -h or --help for more help information")
        sys.exit(1)

    args = parser.parse_args()

    if args.fastq:
        """
        Here generate the code for generating the hhr file first
        """
        P = Parser(args.fastq,
                   args.exp_orientation,
                   args.region_index,
                   args.config_file,
                   args.mpi,
                   args.outputfile,
                   verbose = args.verbose,
                   cleanup = args.cleanup)

    else:
        print("type -h or --help for more help information")
        sys.exit(1)


if __name__ == "__main__":


    Main()



#!/usr/bin/env python3

import lr_tools as lt

import sys
import os
import statistics
import re

from datetime import datetime
import time
import tempfile
import subprocess
import argparse
import pickle
import pprint
import fileinput
from copy import deepcopy
from mpi4py import MPI
from itertools import islice

import fileinput as FI
import h5py

try:
    import configparser
except:
    import ConfigParser as configparser
from configparser import ExtendedInterpolation



class Process:
    def __init__(self, rootname_before, rootname_after, working_dir, count_threshold, nameout, mpi, config_file, pdbfile=False, verbose = False, cleanup = False):
        self.nbef = rootname_before
        self.naft = rootname_after
        self.working_dir = working_dir

        self.nameout = nameout
        self.mpi = mpi
        self.config_file = config_file
        self.log = ""
        self.pdbfile = pdbfile
        self.verbose = verbose
        self.cleanup = cleanup

        # Parameters to pass as options
        self.use_MONO_for_statistics_onWTsilMut = True
        self.use_MONO_for_statistics_onSynMut = True
        self.use_distinct_stats_between_MONO_vs_DITRI = True

        # Analyze count threshold and create a list of threshold for every categories of mutant, MONO, DI or TRI
        if count_threshold:
            if count_threshold == "sum_all":
                self.count_threshold = "sum_all"
            else:
                nthresh = count_threshold.split(',')
                if len(nthresh) == 1:
                    self.count_threshold = [int(count_threshold),int(count_threshold),int(count_threshold)]
                else:
                    self.count_threshold = [int(i) for i in nthresh]
        else:
            self.count_threshold = None


        self.read_config_file()

        self.d3to1, self.d1to3 = lt.get_dicaa()
        self.dmut = lt.parse_twist_devis_table(self.mutfile, self.dn2a, self.d3to1, start=None, stop=None,
                                               info_codon_index=None)
        self.d_enrich = dict()
        self.run_process()

        if self.pdbfile:
            self.map_on_pdb()


    def run_process(self):

        os.chdir(self.working_dir)
        self.comparecond_out = "{}_compare_conditions.out".format(self.nameout)
        self.dmut_cond = {}
        self.mutant_out = "{}_mutations.out".format(self.nameout)
        self.dbcode = {}
        self.dmutcode = {}
        self.dmutcode['order'] = []

        # Dic of all WT silence nuc mutations with counts >0
        self.dwtsilent = {}
        # Dic of all pairs of synonymous mutations with counts >0
        self.dsyno = {}

        for fname_condition in [self.nbef, self.naft]:
            print("Analyzing condition ",fname_condition)
            fbcode = '{}_assembled.out'.format(fname_condition)
            path_fcode = os.path.join(self.working_dir,fbcode)
            mutcode = '{}_mutants_for_rev.out'.format(fname_condition)
            path_mutcode = os.path.join(self.working_dir,mutcode)

            # Analyze the mutant counts for every primer output
            with (open(path_mutcode)) as fin:
                for line in fin:
                    s = line.split()
                    nucmut = s[0]
                    mutation_type = len(nucmut.split('.')) - 1 # 0, 1 or 2 for MONO, DI or TRI
                    aamut = s[1]
                    aalist = [int(s[1][1:-1]),s[1][-1]] # index de mutation, AA mutant
                    if not self.count_threshold:
                        count = eval(s[2])
                        distrib = s[3]
                    elif self.count_threshold == "sum_all":
                        # We sum the entire number of rerads with a given mutation, irrespective of the barcode
                        distrib = s[3]
                        count = sum([int(kk) for kk in distrib.split('_')])
                    else:
                        count, distrib = self.recompute_count_with_thresh(s[3], self.count_threshold[mutation_type])
                    if nucmut not in self.dmutcode:
                        self.dmutcode[nucmut] = {}
                        self.dmutcode['order'].append([aalist[0],aalist[1],nucmut]) # index de mutation AA, AA mutant, nucmut
                    self.dmutcode[nucmut][fname_condition] = [aamut,count,distrib]
                    #print(self.dmutcode[nucmut][fname_condition])
        self.dmutcode['order'].sort()
        #print(self.dmutcode['order'])

        # Get the count values from WT silents
        numberOf_WTenrich, median_WTenrich = self.analyze_silentWT(self.nbef,self.naft)
        print(median_WTenrich)
        # Compare the enrichments between synonymous mutations
        numberOf_Syno, syno_Mean, syno_StdDev = self.analyze_synonymous(self.nbef, self.naft, median_WTenrich)

        self.normalization_factor = median_WTenrich
        self.perturbation_threshold = syno_StdDev

        # Dump the counts into a file
        with(open(self.comparecond_out,"w")) as fout:
            fout.write(self.log)
            fout.write("# COLUMNS = 1) Nucleotide Mutant\n"
                       "#           2) AminoAcid Mutant\n"
                       "#           3) Mutant counts before selection \n"
                       "#           4) Mutant counts after selection \n"
                       "#           5) Enrichment corrected using WTsilent and Synonyms \n")
            fout.write("#\n")

            for mutlist in self.dmutcode['order']:
                mut = mutlist[-1]
                number_of_mutations = len(nucmut.split('.'))
                if number_of_mutations == 1:
                    mutant_category = 0
                else:
                    mutant_category = 1
                if self.use_distinct_stats_between_MONO_vs_DITRI:
                    try:
                        normalization_factor = self.normalization_factor[mutant_category]
                    except IndexError:
                        normalization_factor = self.normalization_factor[0]
                aamut = self.dmutcode[mut][self.nbef][0]
                fout.write("{}\t{}\t".format(mut, aamut))
                dcount = {}
                for fname_condition in [self.nbef, self.naft]:
                    dcount[fname_condition] = self.dmutcode[mut][fname_condition][1]
                    fout.write("{}\t".format(dcount[fname_condition]))
                if dcount[self.nbef] != 0:
                    enrich = dcount[self.naft] * (1./normalization_factor) / dcount[self.nbef]
                    self.d_enrich[mut] = enrich
                    fout.write("{:.2f}\n".format(enrich))
                else:
                    enrich = "Nan"
                    fout.write("{}\n".format(enrich))

    def map_on_pdb(self):
        d_aa2enrich = dict()
        d_aa2pert  = dict()
        d_aa2enrich['order'] = []
        self.pdb_out = "{}_map_perturbation.pdb".format(self.nameout)
        for l_mut in self.dmutcode['order']: # index de mutation AA, AA mutant, nucmut
            if l_mut[-1] not in self.d_enrich:
                continue
            if l_mut[-1] not in self.dmut:
                continue
            enrich = self.d_enrich[l_mut[-1]]
            aamut_index = "{}{}".format(l_mut[1],l_mut[0]) # example A50
            if not aamut_index in d_aa2enrich:
                d_aa2enrich[aamut_index] = []
                d_aa2enrich['order'].append(aamut_index)

            d_aa2enrich[aamut_index].append(enrich)
            print(l_mut,enrich)
        print("perturbation_threshold:",self.perturbation_threshold)
        for aamut in d_aa2enrich['order']:
            mean_enrich =  sum(d_aa2enrich[aamut])*1./len(d_aa2enrich[aamut])
            low_thresh = 1.- self.perturbation_threshold
            high_thresh = 1.+ self.perturbation_threshold
            print(aamut,mean_enrich)
            if aamut[0]=="Z":
                continue
            if mean_enrich > low_thresh and mean_enrich < high_thresh:
                perturbation = 0.
            else:
                perturbation = min(abs(low_thresh-mean_enrich),abs(high_thresh-mean_enrich))
            index = aamut[1:]
            """
            if index=='65' or index=='71': # Excetion for VHH1
                continue
            """
            if perturbation == 0.:
                continue
            if index not in d_aa2pert:
                d_aa2pert[index] = 0
            d_aa2pert[index] += perturbation

        lt.map_perturbation_score2pdb(self.pdbfile, d_aa2pert, self.pdb_out)#,chain_pdb=["G"])

    def analyze_silentWT(self,before,after):
        """
        Return the median of the enrichments for WT silent mutations
        :param before: root_name of the file with mutant counts before selection
        :param after: root_name of the file with mutant counts after selection
        :return:
        """
        # Get the count values from WT silents
        for index,mutaa,nucmut in self.dmutcode['order']:
            number_of_mutations = len(nucmut.split('.'))
            if number_of_mutations == 1:
                mutant_category = 0
            else:
                mutant_category = 1
            if not self.use_MONO_for_statistics_onWTsilMut and number_of_mutations == 1: # We don't consider single mutations due to potential noise
                continue
            aamut,count_bef,distrib = self.dmutcode[nucmut][before]
            aamut,count_aft,distrib = self.dmutcode[nucmut][after]
            #print(aamut, mutaa)
            if aamut[0] == aamut[-1] and count_bef > 0:
                enrich = count_aft*1. / count_bef
                self.dwtsilent[nucmut] = [count_bef, count_aft, enrich, mutant_category]
                #print("WT: {}, {}, {}, {:.3f}".format(nucmut, count_bef, count_aft, enrich))

        list_cases = ["MONO", "DI+TRI"]
        if self.use_distinct_stats_between_MONO_vs_DITRI:
            list_WTenrich = [[] for ii in range(2)]
            list_WTenrich[0] = [self.dwtsilent[i][2] for i in self.dwtsilent if self.dwtsilent[i][3]==0]
            list_WTenrich[1] = [self.dwtsilent[i][2] for i in self.dwtsilent if self.dwtsilent[i][3]==1]
        else:
            list_WTenrich = [[self.dwtsilent[i][2] for i in self.dwtsilent]]
        median_WTenrich = []
        numberOf_WTenrich = []
        cases = []
        for jj in range(len(list_WTenrich)):

            if len(list_WTenrich[jj]) == 0:
                continue
            median_WTenrich.append(statistics.median(list_WTenrich[jj]))
            numberOf_WTenrich.append(len(list_WTenrich[jj]))
            cases.append(list_cases[jj])

        if len(numberOf_WTenrich) > 1:
            message = "{} silent WT mutants {} have a median ratio of counts before and after at {} {}\n"\
                .format(numberOf_WTenrich, cases, median_WTenrich, cases)
        else:
            message = "{} silent WT mutants {} have a median ratio of counts before and after  at {}\n"\
                .format(numberOf_WTenrich, cases, median_WTenrich)
        print("> {}".format(message))
        self.log += "# {}".format(message)

        message += "Ratio of counts will be used as normalization factor to center enrichment values around 1\n"
        print("> {}".format(message))
        self.log += "# {}".format(message)

        return numberOf_WTenrich,median_WTenrich

    def analyze_synonymous(self,before, after, normalization_factor):
        """
        Analyze the standard deviation between the enrichments calculated between synonymous pairs.
        :param before: root_name of the file with mutant counts before selection
        :param after: root_name of the file with mutant counts after selection
        :param normalization: Normalization factor used to compare after and before selections
        :return: standard deviation of the differences between the synonymous pairs.
        """
        # oscar
        print(before, after, normalization_factor)
        set_nucmut_analyzed = set() # set of mutants already seen was soi as not to repeat analysis in symetry

        for index,mutaa,nucmut in self.dmutcode['order']:
            number_of_mutations = len(nucmut.split('.'))
            if number_of_mutations == 1:
                mutant_category = 0
            else:
                mutant_category = 1
            if self.use_distinct_stats_between_MONO_vs_DITRI:
                try:
                    float_normalization_factor = normalization_factor[mutant_category]
                except IndexError:
                    float_normalization_factor = normalization_factor[0]
            else:
                float_normalization_factor = normalization_factor[0]
            if not self.use_MONO_for_statistics_onSynMut and number_of_mutations == 1: # We don't consider single mutations due to potential noise
                continue
            aamut,count_bef,distrib = self.dmutcode[nucmut][before]
            aamut,count_aft,distrib = self.dmutcode[nucmut][after]
            if not self.dmut[nucmut]['nuc_syn']:
                continue
            if count_bef == 0:
                continue
            enrich = count_aft * (1./float_normalization_factor) / count_bef
            for syno in self.dmut[nucmut]['nuc_syn']:
                if not self.use_MONO_for_statistics_onSynMut and number_of_mutations == 1: # We don't consider single mutations due to potential noise
                    continue
                if syno in set_nucmut_analyzed:
                    continue
                number_of_mut_in_syno = len(syno.split('.'))
                if number_of_mut_in_syno == 1:
                    syno_category = 0
                else:
                    syno_category = 1
                if self.use_distinct_stats_between_MONO_vs_DITRI:
                    try:
                        float_normalization_factor2 = normalization_factor[mutant_category]
                    except IndexError:
                        float_normalization_factor2 = normalization_factor[0]
                else:
                    float_normalization_factor2 = normalization_factor[0]
                aamut_syn,count_bef_syn,distrib_syn = self.dmutcode[syno][before]
                aamut_syn,count_aft_syn,distrib_syn = self.dmutcode[syno][after]
                if count_bef_syn == 0:
                    continue
                enrich_syno = count_aft_syn * (1./float_normalization_factor2) / count_bef_syn
                pairsyno_name = "{}_{}&{}".format(aamut,nucmut,syno)
                diff_enrich = enrich - enrich_syno
                self.dsyno[pairsyno_name] = [enrich,enrich_syno,diff_enrich]
            set_nucmut_analyzed.add(nucmut)
        list_diffEnrich = [self.dsyno[i][2] for i in self.dsyno]
        list_couples = [(self.dsyno[i][0],self.dsyno[i][1]) for i in self.dsyno]
        syno_StdDev = statistics.stdev(list_diffEnrich)
        syno_Mean = statistics.mean(list_diffEnrich)
        syno_Length = len(list_diffEnrich)

        message = "{} pairs of synonymous mutations were analyzed\n".format(syno_Length)
        print("> {}".format(message))
        self.log += "# {}".format(message)

        message = "Mean value of the difference between enrichments is : {:2f}\n".format(syno_Mean)
        print("> {}".format(message))
        self.log += "# {}".format(message)

        message = "StdDev value of the difference between enrichments is : {:2f}\n".format(syno_StdDev)
        print("> {}".format(message))
        self.log += "# {}".format(message)

        #pprint.pprint(list_couples)

        return syno_Length,syno_Mean,syno_StdDev

    def recompute_count_with_thresh(self, distrib, thresh):
        list_size = [int(size) for size in distrib.split('_')]
        new_distrib = []
        for elt in list_size:
            if elt >= thresh:
                new_distrib.append(elt)
        return len(new_distrib), '_'.join([str(i) for i in new_distrib])

    def read_config_file(self):
        """

        :return:
        """
        config = configparser.ConfigParser(interpolation=ExtendedInterpolation())
        config.read(self.config_file)

        self.exe_dir    = config.get("Paths", "EXE_DIR")
        self.qqsub_exe  = config.get("Paths", "QSUB_EXE")

        self.use_qqsub   = eval(config.get("Parameters", "USE_QSUB"))

        self.mutfile       = config.get("Files", "TWIST_MUT_TABLE")
        faa2nuc            = config.get("Files", "AA2NUC")
        self.da2n = pickle.load(open(faa2nuc,'rb'),encoding='bytes') # due to py2 to 3
        self.da2n = lt.convert_dic_bytes2string(self.da2n) # due to py2 to 3
        fnuc2aa   = config.get("Files", "NUC2AA")
        self.dn2a = pickle.load(open(fnuc2aa,'rb'))#,encoding='bytes')


def Main():

    try:
        default_configfile = os.path.join(os.path.dirname(os.path.abspath(os.readlink(os.path.abspath(__file__)))),
                                  "NGS_DMS.ini")
    except:
        default_configfile = os.path.join(os.path.dirname(os.path.abspath(os.path.realpath(os.path.abspath(__file__)))),
                                  "NGS_DMS.ini")


    usage = "Script to parse the reads from a long read experiment.\n" + \
            "output is a dictionary encoded as a hdf5 file"
    parser = argparse.ArgumentParser(usage)
    parser.add_argument('-b', '--before_selection', action="store", help='root name <>_assembled.out + <>_mutants_for_rev.out before selection ')
    parser.add_argument('-a', '--after_selection', action="store", help='root name <>_assembled.out + <>_mutants_for_rev.out after selection')
    parser.add_argument('-p', '--pdb_reference', action="store", default=False, help='pdbfile used to map the perturbations')
    parser.add_argument('-d', '--working_directory', action="store", default='./', help='Working directory')
    parser.add_argument('-t', '--threshold', action="store", default=None, help='Which threshold to use for minimal number of counts. '
                                                                                'Default just takes the counts how they are provided')
    parser.add_argument('-o', '--nameout', action="store", default='output', help='output_file')
    parser.add_argument('--mpi', action="store_true", default=False, help='triggers mpi running')
    parser.add_argument('-c', '--config_file', default=default_configfile, action="store", help='name of the configuration file')
    parser.add_argument('--verbose', action="store_true", default=False, help="print extra info")
    parser.add_argument('--cleanup', action="store_true", default=False, help="cleanup temporary files after execution")

    if len(sys.argv) == 1:
        print("type -h or --help for more help information")
        sys.exit(1)

    args = parser.parse_args()

    if args.config_file == default_configfile:
        os.system("cp {} {}".format(default_configfile,os.getcwd()))

    if args.before_selection and args.after_selection :
        """
        Here generate 
        """
        P = Process(args.before_selection,
                      args.after_selection,
                      args.working_directory,
                      args.threshold,
                      args.nameout,
                      args.mpi,
                      args.config_file,
                      pdbfile = args.pdb_reference,
                      verbose = args.verbose,
                      cleanup = args.cleanup)

    else:
        print("type -h or --help for more help information")
        sys.exit(1)

if __name__ == "__main__":


    Main()



#!/usr/bin/env python3

import lr_tools as lt

import sys
import os
import re

from datetime import datetime
import time
import tempfile
import argparse
import pickle
import fileinput
from pprint import pprint
import copy
from mpi4py import MPI
from itertools import islice

import fileinput as FI
import h5py

try:
    import configparser
except:
    import ConfigParser as configparser
from configparser import ExtendedInterpolation



class Assembler:
    def __init__(self, fforward , freverse, fileout, config_file, quality_report, log_file, tag,
                 use_mpi= False, tandem_barcode = False, verbose = False, cleanup = True):
        self.fforward = fforward
        self.freverse = freverse
        self.root_fname = self.get_root_filename(fileout)
        self.config_file = config_file
        self.verbose  = verbose
        self.cleanup  = cleanup
        self.tandem_barcode = tandem_barcode
        self.use_mpi = use_mpi
        if tag:
            self.tag = "_" + tag
        else:
            self.tag = ""

        self.dbarcode = None

        self.thresh_rejection_number_of_observations_per_barcode = 0 # Keep this to 0 to have a fair estimate of
                                                                # the results in case no barcode would have been used.
        self.thresh_maxnumber_of_variants_per_barcode = 6
        self.thresh_number_of_observations_in_multimut_per_barcode = 2
        self.max_interbc_dist = 14 # max distance separating 2 barcodes (7+7, X does not count, dephaser diff => max_dist
        self.thresh_abundance = [(1, 2), (2, 3), (3, 50), (50, 10000)]  # [(1,2)(2,3)(4,10)(10,50)(50,10000)]

        self.quality_report = quality_report
        if self.quality_report:
            print(">>> Quality report activated. Quality report files will be generated.")
            self.dquality = {}
            self.dqcode = {}
            for category in ["W","T","D","S","P"]:
                self.dquality[category] = {}
                self.dquality[category]["low"] = []
                self.dquality[category]["medium"] = []
                self.dquality[category]["high"] = []
                self.dqcode[1] = "S"
                self.dqcode[2] = "D"
                self.dqcode[3] = "T"

        self.d3to1, self.d1to3 = lt.get_dicaa()

        self.start_mpi()

        if self.rank == 0:
            print("\n>>> Reading the configuration file.")

        self.read_config_file()
        self.dmut = self.get_dic_expect_mutations()

        self.run_process()

    def run_process(self):

        #
        # Each read hexad code is assigned a key as a list of properties for each read
        #   [hexa_code] = [variant_code, quality, in_mut_table, barcode]
        #
        number_of_lines_freverse = self.get_nlines(self.freverse)
        number_of_lines_fforward = self.get_nlines(self.fforward)

        self.dreverse,self.lreverse = self.parse_file(self.freverse, number_of_lines_freverse, get_barcode = True)
        if self.rank == 0:
            print("> Size of reads dreverse on every working cpu:",len(self.dreverse))

        self.dforward,self.lforward = self.parse_file(self.fforward, number_of_lines_fforward, get_barcode = self.tandem_barcode)
        if self.rank == 0:
            print("> Size of reads dforward on every working cpu:",len(self.dforward))

        """
        # For NGS_DMS, we need to group together the barcode and decipher the nature of the mutation
        """
        if self.rank==0:
            print(">>> Running self.match_mutant_to_barcodes()")
        dbarcode = self.match_mutant_to_barcodes(fileout = "distrib_barcodes_vs_global_assemby_cases{}.out".format(self.tag))
                # Content of dbarcode key:barcode value:dict(mutant:number_of_reads)
        if self.rank ==0:
            print("> Size of dic of dbarcode:",len(dbarcode))

        # flush memory
        self.dforward = None
        self.dreverse = None
        self.lreverse = None
        self.lforward = None

        dbc_cat = self.sort_barcodes_categories(dbarcode, fileout = "distrib_barcodes_vs_mutant_types{}.out".format(self.tag))

        self.print_to_file_distrib_bc_interdistance(dbc_cat,dir_distrib="distrib_barcode_similarities{}".format(self.tag), fileout="distrib_bcdist{}".format(self.tag))

        # The end of the script runs without mpi

        if self.rank == 0:
            self.print_to_file_compare_bad_correct_mono(dbc_cat, dbarcode, "distrib_bad_mono_in_seq{}.out".format(self.tag), length_protein=(len(self.coding_seq)//3)+2)

            # Eventually not found useful to improve the signal-to-noise ratio
            if self.quality_report:
                print(self.dquality)
                self.analyze_quality()

            self.dobsmutants = self.count_mutants(self.root_fname,dbarcode,code_forw_rev='assembled')
            print("> Number of mutants analyzed :",len(self.dobsmutants))

            # Dump the mutant counts in a file mutant_rev_for.out
            self.analyze_master_mutant()

            """#################
             END OF MAIN SCRIPT
            """#################

    def start_mpi(self):
        if self.use_mpi:
            # MPI communication
            self.comm = MPI.COMM_WORLD
            self.rank = self.comm.Get_rank()
            self.size = self.comm.Get_size()
        else:
            self.comm = None
            self.rank = 0
            self.size = 1

    def get_nlines(self,file):
        if self.rank == 0:
            number_of_lines = lt.count_file_lines(file)
            #print("I finished")
        else:
            number_of_lines = None
        if self.use_mpi:
            self.comm.Barrier()
            number_of_lines = self.comm.bcast(number_of_lines, root=0)
        if self.rank==0:
            print("> The number of reads in the input file is: {}\n".format(number_of_lines))
        return number_of_lines

    def get_root_filename(self,fileout):
        """

        :return:
        """
        s = fileout.split('.')
        if len(s) == 1:
            return fileout
        else:
            return '.'.join(fileout.split('.')[:-1])

    def get_dic_expect_mutations(self):
        """

        :return:
        """
        return lt.parse_twist_devis_table(self.mutfile, self.dn2a, self.d3to1, start=None, stop=None, info_codon_index=None)

    def parse_file(self,file, number_of_lines, get_barcode = False):
        """
        File parsing. Can be used using mpi. Every cpu is taking charge of a chunk of the file.
        This chunk will be passed to the next method so that cpu will work independently on its own chunk
        Only after forward and reverse sets of reads are fused, the barcodes are extracted and mutant assigned.
        The program will then work on the whole library of reads.

        :param file: a one-line for every read file starting with read hexadecimal code, the quality score,
        the nucleotide mutant code and the amino acid mutant code:
        "0x2540be402	0.13	N	12A_42A.43T.44T_46A_100A	AACACAGC"
        Forward and reverse file should  have exactely the same number of lines.
        Since indexing of reads in the file should match exactely between the forward and the reverse file.

        :param get_barcode: extracts the barcode when available.

        :return: Every cpu returns a list of read codes and a dictionary with the following info:
        d[read code] = [variant_code, quality, in_mut_table, barcode]
        """
        # Define the start and end of file reading for every cpu
        size_chunks = (number_of_lines) / self.size + 1 #
        index_start = int(self.rank * (size_chunks))
        if self.rank == self.size - 1:
            index_stop = (number_of_lines)
        else:
            index_stop = int((self.rank + 1) * (size_chunks))

        # Initialization of list and dictionary for every cpu
        if self.use_mpi:
            self.comm.Barrier()
            list_dread2bc = [dict() for i in range(self.size)]
            list_lread2bc = [list() for i in range(self.size)]
            d_read2barcodemut = self.comm.scatter(list_dread2bc,root=0)
            l_read2barcodemut = self.comm.scatter(list_lread2bc,root=0)
            if self.rank==0:
                print(">>> Parsing the file {} using {} cpus".format(file, self.size))
        else:
            print(">>> Parsing the file {} ".format(file))
            d_read2barcodemut = dict()
            l_read2barcodemut = list()

        with open(file) as file_obj:
            n=1
            object_to_iterate = islice(file_obj, index_start, index_stop)
            for line in object_to_iterate:
                if line[0] == "#":
                    continue
                s = line.split()
                read_hexidx  = int(s[0],16)
                quality = int(eval(s[1])*100)
                variant_code = s[3]
                in_mut_table = s[2]
                if get_barcode:
                    try:
                        barcode = s[4]
                    except IndexError:
                        barcode = None
                    d_read2barcodemut[read_hexidx] = [variant_code, quality, in_mut_table, barcode]
                    l_read2barcodemut.append(read_hexidx)
                else:
                    d_read2barcodemut[read_hexidx] = [variant_code, quality, in_mut_table]
                    l_read2barcodemut.append(read_hexidx)
        if self.use_mpi:
            self.comm.Barrier()
            # We don't gather the dictionaries. We return every chunk independently
            """
            d_read2barcodemut = self.comm.gather(d_read2barcodemut, root=0)
            l_read2barcodemut = self.comm.gather(l_read2barcodemut, root=0)
            if self.rank == 0:
                print(">>> Gathered back a list of {} dictionary.".format(self.size))
                for dd in d_read2barcodemut[1:]:
                    d_read2barcodemut[0] = lt.data_merge(d_read2barcodemut[0],dd)
                d_read2barcodemut = d_read2barcodemut[0]
                for ll in l_read2barcodemut[1:]:
                    l_read2barcodemut[0] = lt.data_merge(l_read2barcodemut[0],ll)
                l_read2barcodemut = l_read2barcodemut[0]
                print(">>> Merged them into a single dictionary")
            else:
                l_read2barcodemut = None
                d_read2barcodemut = None
            l_read2barcodemut = self.comm.bcast(l_read2barcodemut, root=0)
            d_read2barcodemut = self.comm.bcast(d_read2barcodemut, root=0)
            if self.rank==0:
                print(">>> Result Broadcast to all cpus")
            """
        return d_read2barcodemut, l_read2barcodemut

    def analyze_quality(self):
        """
        Test to integrate qualities to filter reads but was not kept as an interesting method eventually
        :param tandem_barcodes:
        :return:
        """
        for k in self.dreverse:
            if k in self.dforward:
                forward = self.dforward[k]
                reverse = self.dreverse[k]
                barcode_fow = forward[-1]
                barcode_rev = reverse[-1]
                quality_fow = int(eval(forward[1]) * 100)
                quality_rev = int(eval(reverse[1]) * 100)
                if not (barcode_fow and barcode_rev):  # Both forward and reverse are None
                    self.dquality["P"]["medium"].append(quality_fow)
                    self.dquality["P"]["medium"].append(quality_rev)
                    continue
                # Foward and Reverse Barcodes are merged with an X in the middle
                full_barcode = barcode_fow + "X" + barcode_rev

                mcode_fow = forward[2]
                mcode_rev = reverse[2]

                mutant_fow = forward[0]
                mutant_rev = reverse[0]
                if not full_barcode in self.dquality:
                    self.dquality[full_barcode] = {}
                if mcode_fow + mcode_rev == "OO":
                    # Case 1 : Full agreement between both reads on an expected mutation
                    if mutant_fow == mutant_rev:
                        if not mutant_fow in self.dquality[full_barcode]:
                            self.dquality[full_barcode][mutant_fow] = []
                        self.dquality[full_barcode][mutant_fow].append(quality_fow)
                        self.dquality[full_barcode][mutant_fow].append(quality_rev)
                elif mcode_fow + mcode_rev == "NN":
                    # Case 4: both reads are WT
                    if mutant_fow + mutant_rev == "WW":
                        if not "W" in self.dquality[full_barcode]:
                            self.dquality[full_barcode]["W"] = []
                        self.dquality[full_barcode]["W"].append(quality_fow)
                        self.dquality[full_barcode]["W"].append(quality_rev)
                if len(self.dquality[full_barcode]) == 0:
                    del self.dquality[full_barcode]
            else:
                if self.verbose:
                    print("READ {} NOT FOUND in forward file".format(k))

    def match_mutant_to_barcodes(self, fileout=None, tandem_barcodes = False):
        """
        This method analyzes all the hexacodes for every lines, fuses the barcodes
        and decides the status of the mutation
        Distinguishes 12 categories of mutations

        :param tandem_barcodes: used in a previous version of the script for long read analysis
        :return: a dictionary of barcodes with information about the mutants associated with it.
        """

        if self.rank==0:
            print(">>> Matching forward and reverse reads. Counting barcodes and associated mutations")

        self.list_attributes = ['Expect_mut_agree','WT_agree','WT+Expect_mut',
                                'Expect_DITRI+Bullshit','Expect_mut_disagree_DITRIwin', 'Expect_MONOmut_disagree',
                                'Expect_DITRImut_disagree','ConstantTail_issue','NotExpected',
                                'NotExpect_ContainDIorTRI_retrieved','NotExpect_ContainDIorTRI_notkept',
                                'NotExpect_MONO_kept','NotExpect_DI_kept','NotExpect_TRI_kept']
        dtranslate = {key:value for key, value in [(1,"MONO"),(2,"DI"),(3,"TRI")]}

        len_freads = len(self.lforward)
        len_rreads = len(self.lreverse)
        if len_freads != len_rreads:
            print("ERROR: the number of reads in the forward and the reverse set of reads are not equal. "
                  "Process is stopped."
                  "Check what's going wrong first.")
            sys.exit()

        # Define the start and end of read list analysis for every cpu
        size_chunks = (len_rreads) / self.size + 1 #
        index_start = int(self.rank * (size_chunks))
        if self.rank == self.size - 1:
            index_stop = len_rreads
        else:
            index_stop = int((self.rank + 1) * (size_chunks))

        # Initialisation of the dict
        ### MPI block
        if self.use_mpi:
            self.comm.Barrier()
            if self.rank == 0:
                dbarcode = [dict() for i in range(self.size)]
                count_cases = [{key:0 for key in self.list_attributes} for i in range(self.size)] # Used to perform some statistics
                number_of_common_hexindexes = [0 for i in range(self.size)]
                number_of_unmatched_hexindexes = [0 for i in range(self.size)]
            else:
                dbarcode = None
                count_cases = None
                number_of_common_hexindexes = None
                number_of_unmatched_hexindexes = None
            dbarcode = self.comm.scatter(dbarcode, root=0)
            count_cases = self.comm.scatter(count_cases, root=0)
            number_of_common_hexindexes = self.comm.scatter( number_of_common_hexindexes, root=0)
            number_of_unmatched_hexindexes = self.comm.scatter( number_of_unmatched_hexindexes, root=0)
            if self.rank==0:
                print(">>> Scattered a list of {} dictionary of barcodes and count_cases in different mutant categories".format(self.size))
        else:
            dbarcode = dict()
            count_cases = {key:0 for key in self.list_attributes}
            number_of_common_hexindexes = 0
            number_of_unmatched_hexindexes = 0

        for ii, reverse_code in enumerate(self.lreverse):
            forward_code = self.lforward[ii]
            forward = self.dforward[forward_code] # a list of elements of type d[read code] = [variant_code, quality, in_mut_table, barcode]
            reverse = self.dreverse[reverse_code]
            number_of_common_hexindexes += 1
            barcode_fow = forward[-1]
            barcode_rev = reverse[-1]
            if not (barcode_fow and barcode_rev):  # Either forward or reverse are None
                count_cases['ConstantTail_issue'] += 1
                continue
            # Foward and Reverse Barcodes are merged with an X in the middle
            full_barcode = barcode_fow + "X" + barcode_rev

            # Code is to know if the mutation is canonical or not. Can be O or N
            mcode_fow = forward[2]
            mcode_rev = reverse[2]
            # mutant is the nucleotide mutation
            mutant_fow = forward[0]
            mutant_rev = reverse[0]

            if not full_barcode in dbarcode:
                dbarcode[full_barcode] = {}
            if mcode_fow + mcode_rev == "OO":
                # Case 1 : Full agreement between both reads on an expected mutation
                if mutant_fow == mutant_rev:
                    if not mutant_fow in dbarcode[full_barcode]:
                        dbarcode[full_barcode][mutant_fow] = 0
                    dbarcode[full_barcode][mutant_fow] += 1
                    count_cases['Expect_mut_agree'] += 1
                # Case 2 : Disagreement between expected mutation...
                #          We keep it in the report but not in the enrichments
                else:
                    # Case 2a: One mutation is a di- or tri_ nucleotide variant. The other a mono
                    # Our decision = select the longest variant
                    if len(mutant_fow.split('.')) >= 2 and len(mutant_rev.split('.')) == 1:
                        if not mutant_fow in dbarcode[full_barcode]:
                            dbarcode[full_barcode][mutant_fow] = 0
                        dbarcode[full_barcode][mutant_fow] += 1
                        count_cases['Expect_mut_disagree_DITRIwin'] += 1
                    # Case 2b: One mutation is a di- or tri_ nucleotide variant. The other a mono
                    elif len(mutant_rev.split('.')) >= 2 and len(mutant_fow.split('.')) == 1:
                        if not mutant_rev in dbarcode[full_barcode]:
                            dbarcode[full_barcode][mutant_rev] = 0
                        dbarcode[full_barcode][mutant_rev] += 1
                        count_cases['Expect_mut_disagree_DITRIwin'] += 1
                    # Case 2c: Both mutations are di- or tri_ nucleotide variant. We don't keep.
                    elif len(mutant_rev.split('.')) >= 2 and len(mutant_fow.split('.')) >= 2:
                        mutation = "_".join([mutant_fow, mutant_rev])
                        if not mutation in dbarcode[full_barcode]:
                            dbarcode[full_barcode][mutation] = 0
                        dbarcode[full_barcode][mutation] += 1
                        count_cases['Expect_DITRImut_disagree'] += 1
                    # Case 2d:  Both mutations are mono nucleotide variant. We don't keep.
                    else:
                        mutation = "_".join([mutant_fow, mutant_rev])
                        if not mutation in dbarcode[full_barcode]:
                            dbarcode[full_barcode][mutation] = 0
                        dbarcode[full_barcode][mutation] += 1
                        count_cases['Expect_MONOmut_disagree'] += 1
            elif mcode_fow + mcode_rev in ["ON","NO"]:
                # Case 3: One side was seen WT while the other was seen mutated as expected
                if mutant_rev == "W":
                    # 3a : Mutation is in the forward
                    if not mutant_fow in dbarcode[full_barcode]:
                        dbarcode[full_barcode][mutant_fow] = 0
                    dbarcode[full_barcode][mutant_fow] += 1
                    count_cases['WT+Expect_mut'] += 1
                elif mutant_fow == "W":
                    # 3b : Mutation is in the reverse
                    if not mutant_rev in dbarcode[full_barcode]:
                        dbarcode[full_barcode][mutant_rev] = 0
                    dbarcode[full_barcode][mutant_rev] += 1
                    count_cases['WT+Expect_mut'] += 1
                else:
                    # 3c : Mutation is in the forward (>=2 means DI or TRI). The other is bullshit
                    # We tested >=1 but std_dev of synonymous increased
                    if mcode_fow == "O" and len(mutant_fow.split('.')) >= 2:
                        if not mutant_fow in dbarcode[full_barcode]:
                            dbarcode[full_barcode][mutant_fow] = 0
                        dbarcode[full_barcode][mutant_fow] += 1
                        count_cases['Expect_DITRI+Bullshit'] += 1
                    # 3d : Mutation is in the reverse (>=2 means DI or TRI). The other is bullshit
                    # We tested >=1 but std_dev of synonymous increased
                    elif mcode_rev == "O" and len(mutant_rev.split('.')) >= 2:
                        if not mutant_rev in dbarcode[full_barcode]:
                            dbarcode[full_barcode][mutant_rev] = 0
                        dbarcode[full_barcode][mutant_rev] += 1
                        count_cases['Expect_DITRI+Bullshit'] += 1
            elif mcode_fow + mcode_rev == "NN":
                # Case 4: both reads are WT
                if mutant_fow + mutant_rev == "WW":
                    if not "WW" in dbarcode[full_barcode]:
                        dbarcode[full_barcode]["WW"] = 0
                    dbarcode[full_barcode]["WW"] += 1
                    count_cases['WT_agree'] += 1
                # Case 5: Reads do not contain expected sequence
                else:
                    count_cases['NotExpected'] += 1
                    possible_fow_mut_partial = []
                    possible_rev_mut_partial = []
                    # We try to save some mutants by detecting those DI or TRI which could be canonical
                    # but spoiled by a single perturbation.
                    if len(mutant_fow.split('_')) == 2:
                        for mut_partial in mutant_fow.split('_'):
                            if len(mut_partial.split('.')) >=2 and mut_partial in self.dmut:
                                possible_fow_mut_partial.append(mut_partial)
                    if len(mutant_rev.split('_')) == 2:
                        for mut_partial in mutant_rev.split('_'):
                            if len(mut_partial.split('.')) >=2 and mut_partial in self.dmut:
                                possible_rev_mut_partial.append(mut_partial)
                    mutant_select = None
                    if len(possible_rev_mut_partial) == 1 and len(possible_fow_mut_partial)==1:
                        if possible_rev_mut_partial[0] == possible_fow_mut_partial[0]:
                            mutant_select = possible_rev_mut_partial[0]
                    if len(possible_rev_mut_partial) == 1 and len(possible_fow_mut_partial)==0:
                        mutant_select = possible_rev_mut_partial[0]
                    if len(possible_rev_mut_partial) == 0 and len(possible_fow_mut_partial)==1:
                        mutant_select = possible_fow_mut_partial[0]
                    if mutant_select:
                        if not mutant_select in dbarcode[full_barcode]:
                            dbarcode[full_barcode][mutant_select] = 0
                        dbarcode[full_barcode][mutant_select] += 1
                        count_cases['NotExpect_ContainDIorTRI_retrieved'] += 1
                    else:
                        count_cases['NotExpect_ContainDIorTRI_notkept'] += 1

                    # We keep the single mut MONO that are not in self.dmut for statistical analysis
                    if len(mutant_fow.split('_')) == 1 and len(mutant_rev.split('_')) == 1:
                        if (len(mutant_fow.split('.')) in [1,2,3] or mutant_fow == "W") and (len(mutant_rev.split('.')) in [1,2,3] or mutant_rev == "W"):
                            if mutant_fow == mutant_rev or mutant_fow == "W" or mutant_rev == "W":
                                if mutant_fow == "W":
                                    mutant_select = mutant_rev
                                elif mutant_rev == "W":
                                    mutant_select = mutant_fow
                                else:
                                    mutant_select = mutant_fow
                                if not mutant_select in dbarcode[full_barcode]:
                                    dbarcode[full_barcode][mutant_select] = 0
                                dbarcode[full_barcode][mutant_select] += 1
                                count_cases['NotExpect_{}_kept'.format(dtranslate[len(mutant_select.split('.'))])] += 1
            if len(dbarcode[full_barcode]) == 0:
                del dbarcode[full_barcode]

        ### BLOCK MPI
        if self.use_mpi:
            self.comm.Barrier()
            dbarcode = self.comm.gather(dbarcode, root=0)
            count_cases = self.comm.gather(count_cases, root=0)
            number_of_common_hexindexes = self.comm.gather(number_of_common_hexindexes, root=0)
            number_of_unmatched_hexindexes = self.comm.gather(number_of_unmatched_hexindexes, root=0)

            if self.rank == 0:
                number_of_common_hexindexes = sum(number_of_common_hexindexes)
                number_of_unmatched_hexindexes = sum(number_of_unmatched_hexindexes)
                print("> Gathered back the list of {} dictionary.".format(self.size))
                for jj, dd in enumerate(dbarcode[1:]):
                    dbarcode[0] = lt.data_merge(dbarcode[0],dd,sum_values = True)
                dbarcode = dbarcode[0]
                for dd in count_cases[1:]:
                    count_cases[0] = lt.data_merge(count_cases[0],dd,sum_values = True)
                count_cases = count_cases[0]

            dbarcode = self.comm.bcast(dbarcode, root=0)

        if self.rank == 0:
            if fileout:
                fout = open(fileout,"w")
                fout.write("# Statistics comparing forward and reverse read\n")
                fout.write("# Total number of matched pair of indexes = {}\n".format(number_of_common_hexindexes))
                fout.write("# Total number of unmatched pair of indexes = {}\n".format(number_of_unmatched_hexindexes))
                total = 0
                for elt in self.list_attributes:
                    fout.write("{} :\t{}\n".format(elt,count_cases[elt]))
                    if elt not in ['NotExpect_ContainDIorTRI_retrieved','NotExpect_ContainDIorTRI_notkept','NotExpect_MONO_kept','NotExpect_DI_kept','NotExpect_TRI_kept']:
                        total += count_cases[elt]
                fout.write("Unassigned :\t{}\n".format(number_of_common_hexindexes-total))
                fout.write("Total of assigned reads in categories :\t{}\n".format(total))

        return dbarcode

    def count_mutants(self,fileoutroot, dbarcode, code_forw_rev='assembled'):
        """
        Analyzes through the dbarcode dictionary
        Writes the barcodes in a file with their first hits.
        :return: The dictionary dmut which associates every nucmut with a list of counts
        """
        if self.rank == 0:
            print(">>> Analyzing for every mutant, the number of barcodes.")
        dmut = {}
        fout_name = '{}_{}.out'.format(fileoutroot,code_forw_rev)
        with open(fout_name,"w") as fout:
            for bc in dbarcode:
                list_mutant = []
                for mut in dbarcode[bc]:
                    list_mutant.append([dbarcode[bc][mut],mut])
                # Mutants associated to every barcodes are sorted with respect to their number of observations
                list_mutant.sort(reverse=True) # of the form [[ncount1,mut1], [ncount2,mut2], ...]
                # Barcodes for which the max number of observation is below threshold are not analyzed further
                if list_mutant[0][0] <= self.thresh_rejection_number_of_observations_per_barcode:
                    continue
                fout.write("{}\t".format(bc))
                for ii,m in enumerate(list_mutant):
                    if ii >= 4:
                        break
                    if m[1] == "W":
                        canonical = "W"
                    elif m[1] in self.dmut:
                        canonical = "O"
                    else:
                        canonical = "N"
                    fout.write("{} {} {} | ".format(m[0],canonical,m[1]))
                    # If ii == 0 We only take into account the first mutant associated with a barcode
                    # In case we have mutiple mutants per barcode
                    if ii < self.thresh_maxnumber_of_variants_per_barcode:
                        if canonical in set(["W", "O"]):
                            if ii != 0 and m[0] < self.thresh_number_of_observations_in_multimut_per_barcode:
                                # If the additional mutants are below the threshold we don't keep them
                                # Multi mutants are only considered if they are numerous
                                break
                            if not m[1] in dmut:
                                dmut[m[1]] = []
                            dmut[m[1]].append(m[0])
                fout.write("\n")
        return dmut

    def sort_barcodes_categories(self, dbarcode, fileout = None ):
        """
        Classify barcodes in different categories depending on the type of mutants and the abundance of reads associated
        Writes the barcodes in a file with their first hits.
        :return: The dictionary dbc_cat which associates a set of bacodes with every mutant categories
        """
        # Define the start and end of bc list analysis for every cpu
        size_chunks = (len(dbarcode)) / self.size + 1 #
        index_start = int(self.rank * (size_chunks))
        if self.rank == self.size - 1:
            index_stop = len(dbarcode)
        else:
            index_stop = int((self.rank + 1) * (size_chunks))


        mutcat = ["WT","MONO","DI","TRI","BAD.MONO","BAD.DI","BAD.TRI"]

        # Initialisation of the dict
        dbc_cat_ori = {key:{"{}-{}".format(ini,end):set() for ini,end in self.thresh_abundance} for key in mutcat}
        if self.rank == 0: # Very important otherwise every list won't be identical => BIG MESS
            bc_list = list(dbarcode.keys())
        else:
            bc_list = None

        ### MPI block
        if self.use_mpi:
            self.comm.Barrier()
            bc_list = self.comm.bcast(bc_list,root=0)
            list_of_dbc_cat = [dict(dbc_cat_ori) for i in range(self.size)]
            dbc_cat = self.comm.scatter(list_of_dbc_cat, root=0)
            if self.rank==0:
                print(">>> Scattered a list of {} dictionary of barcodes lists in different mutant categories".format(self.size))
        else:
            dbc_cat = dbc_cat_ori

        if self.verbose:
            print("I'm rank {} and I will work of bc_list from {} to {}".format(self.rank,index_start,index_stop))

        for bc in bc_list[index_start:index_stop]:
            list_mutant = []
            for mut in dbarcode[bc]:
                list_mutant.append([dbarcode[bc][mut],mut])
            # Mutants associated to every barcodes are sorted with respect to their number of observations
            list_mutant.sort(reverse=True)
            # Barcodes for which the max number of observation is below threshold are not analyzed further
            if list_mutant[0][0] <= self.thresh_rejection_number_of_observations_per_barcode:
                continue
            for ii,m in enumerate(list_mutant):
                #if bc == "CGCCCCTCGCXGTGAGAAGCG":
                #    print("DEBUG",ii,m,list_mutant)
                if ii >= self.thresh_maxnumber_of_variants_per_barcode:
                    break
                if m[1] == "WW":
                    for ii,threshs in enumerate(self.thresh_abundance):
                        key_thresh = "{}-{}".format(threshs[0],threshs[1])
                        if m[0] >= threshs[0] and m[0] < threshs[1]:
                            #if bc == "CGCCCCTCGCXGTGAGAAGCG":
                            #    print("DEBUG",ii,m,'WW',key_thresh)

                            dbc_cat["WT"][key_thresh].add(bc) # ii is the range of abundance
                else:
                    mut_type = '.'.join(lt.get_mutation_type(m[1],self.dmut)) # e.g. to define BAD.MONO or MONO.
                    if mut_type in mutcat:
                        for ii,threshs in enumerate(self.thresh_abundance):
                            key_thresh = "{}-{}".format(threshs[0], threshs[1])
                            if m[0] >= threshs[0] and m[0] < threshs[1]:
                                #if bc == "CGCCCCTCGCXGTGAGAAGCG":
                                #    print("DEBUG",ii,m,mut_type,key_thresh)

                                dbc_cat[mut_type][key_thresh].add(bc) # ii is the range of abundance

        ### BLOCK MPI
        if self.use_mpi:
            self.comm.Barrier()
            dbc_cat = self.comm.gather(dbc_cat, root=0)
            if self.rank == 0:
                print("> Gathered back the list of {} dictionary.".format(self.size))
                for dd in dbc_cat[1:]:
                    dbc_cat[0] = lt.data_merge(dbc_cat[0],dd)
                dbc_cat = dbc_cat[0]
            dbc_cat = self.comm.bcast(dbc_cat, root=0)

        # Print the distribution to an output file
        if self.rank == 0:
            if fileout:
                fout = open(fileout, "w")
                fout.write("# Number of barcodes for different categories of mutants found with a certain abundance\n")
                fout.write("# $1 = Categories of mutants\n")
                fout.write("# $2 = Range of read abundance first_value <= abundance < second_value\n")
                fout.write("# $3 = Numbers of barcodes\n")
                for cat in mutcat:
                    for ii,thresh in enumerate(self.thresh_abundance):
                        key_thresh = "{}-{}".format(thresh[0], thresh[1])
                        report = "{}\t{}-{}\t{}\n".format(cat,thresh[0],thresh[1], len(dbc_cat[cat][key_thresh]))
                        fout.write("{}".format(report))
                fout .close()

        return dbc_cat

    def analyze_master_mutant(self):
        """
        Generates the file in which for each mutant we retrieve the distribution of the weighted barcodes
        :return:
        """
        fout_name = self.root_fname + '_mutants_for_rev.out'
        d_obs_mutants = [self.dobsmutants]
        files = [fout_name]
        for ii in range(len(files)):
            with open(files[ii],"w") as fout:
                for mut_nuc_name in self.dmut['order']: # We thread through all the expected mutants of the experiment
                    aa_mutant_name = self.dmut[mut_nuc_name]['full_mut_AA_name'].split("@@")
                    if mut_nuc_name in d_obs_mutants[ii]:
                        distrib_largest = list(d_obs_mutants[ii][mut_nuc_name])
                        distrib_largest.sort(reverse=True)
                        str_distrib_largest = '_'.join([str(x) for x in distrib_largest[:]])
                        mutant_number = len(d_obs_mutants[ii][mut_nuc_name])
                        fout.write("{}\t{}\t{}\t{}\n".format(mut_nuc_name, aa_mutant_name[1],mutant_number, str_distrib_largest))
                    else:
                        fout.write("{}\t{}\t0\t0\n".format(mut_nuc_name, aa_mutant_name[1]))

    def print_to_file_distrib_bc_interdistance(self, dbc_cat, dir_distrib = "distrib_barcode_similarities", fileout="distrib_bcdist"):
        """
        :param dbc_cat: dictionnary of categories pointing to sets of barcodes contained in each
        categories are typically : [MONO, BAD.MONO, TRI, etc...][number_of_reads supported by the barcode]
        :param dir_distrib:
        :param fileout:
        :return:
        """
        # indexes of dbc_cat correspond to the number of reads category
        category_to_print = [] # contain 3 elements:
        # 1 - a set of bc to analyse,
        # 2 - a list of sets to compare with,
        # 3 - Boolean to assess whether the bc to analyze should be by default excluded from the refset because it is contained in the reference
        category_to_print.append(["BAD.MONO",["WT","MONO"]])
        category_to_print.append(["BAD.DI",["MONO","DI"]])
        category_to_print.append(["BAD.TRI",["DI","TRI"]])
        category_to_print.append(["MONO",["WT","MONO"]])
        category_to_print.append(["DI",["MONO","DI"]])
        category_to_print.append(["TRI",["DI","TRI"]])
        #dir_distrib = "distrib_barcode_similarities{}".format(self.tag)
        if self.use_mpi:
            self.comm.Barrier()
        if self.rank == 0 and not os.path.isdir(dir_distrib):
            os.mkdir(dir_distrib)
        for threshs in [(2, 3), (50, 10000)]:
            ref_thresh = "{}-{}".format(self.thresh_abundance[-1][0],self.thresh_abundance[-1][1])
            test_thresh = "{}-{}".format(threshs[0],threshs[1])
            if test_thresh == ref_thresh:
                exclude_bc_from_refset = True
            else:
                exclude_bc_from_refset = False
            for cat in category_to_print:
                if cat[0][0] == "B" : # if it is a BAD mutant we don't have to remove it from the ref bc provided they are not BAD
                    exclude_bc_from_refset = False
                set2test = copy.copy(dbc_cat[cat[0]][test_thresh])
                setref = copy.copy(dbc_cat[cat[1][0]][ref_thresh]) # We combine several sets if required in the ref set
                for mutcat in cat[1][1:]:
                    setref.update(dbc_cat[mutcat][ref_thresh])
                filename = os.path.join(dir_distrib, "{}_{}_{}_vs_{}_{}.out".format(fileout, cat[0],test_thresh,"x".join(cat[1]),ref_thresh))
                if self.rank == 0:
                    print(">>> Running compare_barcodes_similarities on {}_{}_vs_{}_{}".format(cat[0],test_thresh,"x".join(cat[1]),ref_thresh))
                self.compare_barcodes_similarities(set2test, setref,filename, exclude_itself = exclude_bc_from_refset)

    def compare_barcodes_similarities(self, set_barcodes2search, set_refbarcodes, fileout, exclude_itself=False):
        """
        Scans a list of barcodes and compare it to a second list of barcodes to define for each barcode of the first
        ensemble what is the minimal divergence to one of the barcode for the second

        :param set_barcodes2search: The first set of barcodes
        :param set_refbarcodes: The second set of barcodes
        :param fileout: name of the file in which solely the distribution of the minimum divergence will be dumped
        :return: Histogram of the distributions
        """

        # Define the start and end of bc list analysis for every cpu
        size_chunks = (len(set_barcodes2search)) / self.size + 1 #
        index_start = int(self.rank * (size_chunks))
        if self.rank == self.size - 1:
            index_stop = len(set_barcodes2search)
        else:
            index_stop = int((self.rank + 1) * (size_chunks))

        # Defines a unique list of ordered barcodes. Required for chunking
        if self.rank == 0:
            list_barcodes2search = list(set_barcodes2search)
        else:
            list_barcodes2search = None

        ### BLOCK MPI
        if self.use_mpi:
            self.comm.Barrier()
            list_barcodes2search = self.comm.bcast(list_barcodes2search,root=0)
            list_of_ldminid = [[] for i in range(self.size)]
            l_min_id = self.comm.scatter(list_of_ldminid, root=0)
            if self.rank==0:
                print(">>> Scattered a list of {} dictionary of {} barcodes lists in different mutant categories".format(self.size,len(l_min_id)))
        else:
            l_min_id = []

        if self.verbose:
            print("I'm rank {} and I will work on list_barcodes2search from {} to {}".format(self.rank,index_start,index_stop))

        for bc in list_barcodes2search[index_start:index_stop]:
            min_id = lt.compute_minimal_id_with_bcref(bc,set_refbarcodes, exclude_bc = exclude_itself)
            l_min_id.append(min_id)

        ### BLOCK MPI
        if self.use_mpi:
            self.comm.Barrier()
            l_min_id = self.comm.gather(l_min_id, root=0)
            if self.rank == 0:
                print(">>> Gathered back the list of {} dictionary.".format(self.size))
                for ll in l_min_id[1:]:
                    l_min_id[0] = lt.data_merge(l_min_id[0],ll)
                l_min_id = l_min_id[0]
        if self.rank == 0:
            fout = open(fileout,"w")
            fout.write("# For every inter-barcode distance category (col. 1), number of cases are reported (col. 2)\n")
            for dist in range(self.max_interbc_dist+1):
                try:
                    count_dist = l_min_id.count(dist)
                except:
                    count_dist = 0
                fout.write("{}\t{}\n".format(dist,count_dist))
            fout.close()

    def print_to_file_compare_bad_correct_mono(self, dbc_cat, dbarcode, fileout, length_protein=200):
        """
        :param dbc_cat:
        :param fileout:
        :param length_protein:
        :return:
        """
        fout2 = open(fileout ,"w")
        cat2analyze = ["MONO","BAD.MONO"]
        # in dicindex, first 0 is for aa, the three next are for nucleotide count
        dicindex = {key:[[0,0,0,0] for kk in range(length_protein)] for key in cat2analyze}
        for cat in cat2analyze:
            for bc in dbc_cat[cat]["50-10000"]:
                list_mut_with_bc = [[val,key] for key,val in dbarcode[bc].items()]
                list_mut_with_bc.sort(reverse=True)
                #print("BC_TEST",bc)
                for count,mutant in list_mut_with_bc:
                    #print("MUTANT_TEST", mutant, count)
                    len_mutaa = len(mutant.split('_'))
                    len_mutnuc = len(mutant.split('.'))
                    if len_mutnuc == 1 and len_mutaa == 1:
                        if cat == 'MONO' and mutant not in self.dmut:
                            continue
                        if cat == 'BAD.MONO' and mutant in self.dmut:
                            continue
                        list_aamut, list_nucmut = lt.get_aamut(self.dn2a, self.coding_seq.strip(), mutant)
                        if not list_aamut:
                            continue
                        indexnuc = int(mutant[:-1])%3
                        aamut = list_aamut[0]
                        indexaa = int(aamut[1:-1])
                        dicindex[cat][indexaa][0] += 1
                        dicindex[cat][indexaa][1+indexnuc] += 1
                        #if cat == 'MONO' and mutant in self.dmut:
                        #    print('MONO', mutant, list_aamut, bc, dicindex[cat][indexaa])
                        break
        for ii in range(length_protein):
            fout2.write("{}\t".format(ii))
            for cat in cat2analyze:
                index = '_'.join((str(kk) for kk in dicindex[cat][ii]))
                fout2.write("{}\t".format(index))
            fout2.write("\n")
        fout2.close()

    def compare_reverse_vs_forward(self):
        """
        Not used anymore
        :return:
        """
        dbarcode = {}
        for ii,k in enumerate(self.dreverse):
            if k in self.dforward:
                forward = self.dforward[k]
                reverse = self.dreverse[k]
                if forward[-1] == reverse[-1]: # compare the two barcodes
                    barcode = forward[-1]
                    if not barcode: # Both forward and reverse are None
                        continue
                    if not barcode in dbarcode:
                        dbarcode[barcode] = {}
                    mutation = forward[0]
                    if not mutation in dbarcode[barcode]:
                        dbarcode[barcode][mutation] = 0
                    dbarcode[barcode][mutation] += 1
            else:
                if self.verbose:
                    print("READ {} NOT FOUND in forward file".format(k))
        return dbarcode

    def dump_barcode_dictionary(self):
        """
        Not used anymore
        Writes a file listing all the barcodes.
        :return:
        """
        dmut = {}
        with open(self.fileout,"w") as fout:
            for bc in self.fdbarcode:
                list_mutant = []
                for mut in self.fdbarcode[bc]:
                    list_mutant.append([self.fdbarcode[bc][mut],mut])
                list_mutant.sort(reverse=True) # list mutant is like [[ncount1,mutname1],[ncount2,mutname2],...]
                if list_mutant[0][0] <= self.thresh_rejection_number_of_observations_per_barcode:
                    continue
                fout.write("{}\t".format(bc))
                for ii,m in enumerate(list_mutant):
                    if ii >= 4:
                        break
                    if m[1] == "W":
                        canonical = "W"
                    elif m[1] in self.dmut:
                        canonical = "O"
                    else:
                        canonical = "N"
                    fout.write("{} {} {} | ".format(m[0],canonical,m[1]))
                    if ii == 0:
                        if canonical in set(["W","O"]):
                            if not m[1] in dmut:
                                dmut[m[1]] = set()
                            dmut[m[1]].add(m[0])
                fout.write("\n")
        return dmut

    def dump_quality(self):
        thresh_low = 5
        thresh_medium = 50
        for bc in self.dbarcode:
            if bc in self.dquality:
                for mut in self.dbarcode[bc].keys():
                    if mut == "WW":
                        for qual in self.dquality[bc]["W"]:
                            if self.dbarcode[bc][mut] < thresh_low:
                                self.dquality["W"]["low"].append(qual)
                            elif self.dbarcode[bc][mut] >= thresh_low and self.dbarcode[bc][mut] < thresh_medium:
                                self.dquality["W"]["medium"].append(qual)
                            elif self.dbarcode[bc][mut] >= thresh_medium:
                                self.dquality["W"]["high"].append(qual)
                    elif mut in self.dquality[bc]:
                        mut_type = self.dqcode[len(mut.split('.'))] # S, D, T
                        for qual in self.dquality[bc][mut]:
                            if self.dbarcode[bc][mut] < thresh_low:
                                self.dquality[mut_type]["low"].append(qual)
                            elif self.dbarcode[bc][mut] >= thresh_low and self.dbarcode[bc][mut] < thresh_medium:
                                self.dquality[mut_type]["medium"].append(qual)
                            elif self.dbarcode[bc][mut] >= thresh_medium:
                                self.dquality[mut_type]["high"].append(qual)
        for category in ["W","P","S","D","T"]:
            for elt in self.dquality[category]:
                fout = open("quality_{}_{}_{}.out".format(self.root_fname, category, elt),"w")
                for qual in self.dquality[category][elt]:
                    fout.write("{}\n".format(qual))
                fout.close()


    def read_config_file(self):
        """

        :return:
        """
        config = configparser.ConfigParser(interpolation=ExtendedInterpolation())
        config.read(self.config_file)

        self.coding_seq    = lt.get_seq_from_fasta(config.get("Files", "CODING_SEQ"))
        #print("CODING_SEQUENCE:\n{}".format(self.coding_seq))
        self.amplic_seq    = lt.get_seq_from_fasta(config.get("Files", "AMPLICON_SEQ"))

        self.mutfile       = config.get("Files", "TWIST_MUT_TABLE")
        faa2nuc            = config.get("Files", "AA2NUC")
        self.da2n = pickle.load(open(faa2nuc,'rb'),encoding='bytes') # due to py2 to 3
        self.da2n = lt.convert_dic_bytes2string(self.da2n) # due to py2 to 3
        fnuc2aa   = config.get("Files", "NUC2AA")
        self.dn2a = pickle.load(open(fnuc2aa,'rb'))#,encoding='bytes')

        mcodseq2ampli = re.search(self.coding_seq, self.amplic_seq)
        self.start_codseq_in_ampliseq, self.stop_codseq_in_ampliseq = mcodseq2ampli.span()


def Main():

    try:
        default_configfile = os.path.join(os.path.dirname(os.path.abspath(os.readlink(os.path.abspath(__file__)))),
                                  "NGS_DMS.ini")
    except:
        default_configfile = os.path.join(os.path.dirname(os.path.abspath(os.path.realpath(os.path.abspath(__file__)))),
                                  "NGS_DMS.ini")


    usage = "Script to parse the reads from a long read experiment.\n" + \
            "output is a dictionary encoded as a hdf5 file"
    parser = argparse.ArgumentParser(usage)
    parser.add_argument('-f', '--forward_parsed', action="store", help='input forward file')
    parser.add_argument('-r', '--reverse_parsed', action="store", help='input reverse file')
    parser.add_argument('-b', '--tandem_barcode', action="store_true", help='if barcode is present in forward and reverse')
    parser.add_argument('-o', '--root_outfilename', action="store", help='output barcode analysis file')
    parser.add_argument('-q', '--quality_report', action="store_true", default=False,  help='Should a Phred quality report be generated over the different categories')
    parser.add_argument('-c', '--config_file', default=default_configfile, action="store", help='name of the configuration file')
    parser.add_argument('-l', '--log_file', default=False, action="store", help='name of the log file')
    parser.add_argument('-t', '--tag', default=None, action="store", help='string used to specify the output filenames')
    parser.add_argument('--mpi', action="store_true", default=False, help='triggers mpi running')
    parser.add_argument('--verbose', action="store_true", default=False, help="print extra info")
    parser.add_argument('--cleanup', action="store_true", default=False, help="cleanup temporary files after execution")

    if len(sys.argv) == 1:
        print("type -h or --help for more help information")
        sys.exit(1)

    args = parser.parse_args()

    if args.forward_parsed and args.reverse_parsed:
        """
        Here generate the code for generating the hhr file first
        """
        A = Assembler(args.forward_parsed,
                      args.reverse_parsed,
                      args.root_outfilename,
                      args.config_file,
                      args.quality_report,
                      args.log_file,
                      args.tag,
                      use_mpi=args.mpi,
                      tandem_barcode = args.tandem_barcode,
                      verbose = args.verbose,
                      cleanup = args.cleanup)

    else:
        print("type -h or --help for more help information")
        sys.exit(1)


if __name__ == "__main__":


    Main()
